'''
:copyright:
    The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
    Christoph Sens-Schönefelder
    Peter Makus (makus@gfz-potsdam.de)

Created: Monday, 14th June 2021 08:50:57 am
Last Modified: Monday, 16th January 2023 11:13:58 am
'''

from typing import List, Tuple, Optional
import numpy as np
from copy import deepcopy
from scipy.signal import butter, lfilter, hilbert, resample
from scipy.interpolate import UnivariateSpline, interp1d
from obspy.core import UTCDateTime

# Obspy imports
from obspy.signal.invsim import cosine_taper

from seismic.monitor.stretch_mod import multi_ref_vchange_and_align, \
    time_shift_estimate, compare_with_modified_reference, \
    create_shifted_ref_mat
from seismic.correlate.stats import CorrStats
from seismic.monitor.dv import DV
from seismic.monitor.trim import corr_mat_trim


def corr_mat_clip(A: np.ndarray, thres: float, axis: int) -> np.ndarray:
    """
    Clip the Input Matrix upper and lower bounds to a multiple of its
    standard deviation. `thres` determines the factor and `axis` the axis
    the std should be computed over

    :param A: Input Array to be clipped
    :type A: np.ndarray
    :param thres: factor of the standard deviation to clip by
    :type thres: float
    :param axis: Axis to compute the std over and, subsequently clip over.
        Can be None, if you wish to compute floating point rather than a
        vector. Then, the array will be clipped evenly.
    :type axis: int
    :return: The clipped array
    :rtype: np.ndarray
    """
    clipv = thres * np.std(A, axis=axis)
    if axis == 1:
        A = np.clip(A.T, -clipv, clipv).T
    else:
        A = np.clip(A, -clipv, clipv)
    return A


def _smooth(
    x: np.ndarray, window_len: int = 10,
        window: str = 'hanning') -> np.ndarray:
    """ Smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    :type x: :class:`~numpy.ndarray`
    :param x: the input signal
    :type window_len: int
    :param window_len: the dimension of the smoothing window
    :type window: string
    :param window: the type of window from 'flat', 'hanning', 'hamming',
        'bartlett', 'blackman' flat window will produce a moving average
        smoothing.

    :rtype: :class:`~numpy.ndarray`
    :return: The smoothed signal

    >>>import numpy as np
    >>>t = np.linspace(-2,2,0.1)
    >>>x = np.sin(t)+np.random.randn(len(t))*0.1
    >>>y = smooth(x)

    .. rubric:: See also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman,
    numpy.convolve, scipy.signal.lfilter

    TODO: the window parameter could be the window itself if it is an array
        instead of a string
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if window not in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming',\
            'bartlett', 'blackman'")

    s = np.r_[2 * x[0] - x[window_len:1:-1], x,
              2 * x[-1] - x[-1:-window_len:-1]]

    if window == 'flat':  # moving average
        w = np.ones(window_len, 'd')
    else:
        w = getattr(np, window)(window_len)
    y = np.convolve(w / w.sum(), s, mode='same')

    return y[window_len - 1:-window_len + 1]


def corr_mat_smooth(
    data: np.ndarray, wsize: int, wtype: str = 'flat',
        axis: int = 1) -> np.ndarray:
    """ Smoothing of a correlation matrix.

    Smoothes the correlation matrix with a given window function of the given
    width along the given axis. This method is based on the convolution of a
    scaled window with the signal. Each row/col (i.e. depending on the selected
    ``axis``) is "prepared" by introducing reflected copies of it (with the
    window size) in both ends so that transient parts are minimized in the
    beginning and end part of the resulting array.

    :type corr_mat: dictionary of the type correlation matrix
    :param corr_mat: correlation matrix to be smoothed
    :type wsize: int
    :param wsize: Window size
    :type wtype: string
    :param wtype: Window type. It can be one of:
            ['flat', 'hanning', 'hamming', 'bartlett', 'blackman'] defaults to
            'flat'
    :type axis: int
    :param axis: Axis along with apply the filter. O: smooth along correlation
              lag time axis 1: smooth along time axis

    :rtype: :class:`~numpy.ndarray`
    :return: **X**: Filtered matrix
    """
    # Remove nans
    np.nan_to_num(data, copy=False)

    # Degenerated corr_mat: single vector
    try:
        row, col = data.shape
    except ValueError:
        # Single vector not a matrix
        data = _smooth(data, window_len=wsize, window=wtype)
        return data

    # Proper 2d matrix. Smooth on the chosen axis
    if axis == 0:
        for i in np.arange(row):
            csig = data[i]
            scsig = _smooth(csig, window_len=wsize, window=wtype)
            data[i] = scsig
    elif axis == 1:
        for i in np.arange(col):
            csig = data[:, i]
            scsig = _smooth(csig, window_len=wsize, window=wtype)
            data[:, i] = scsig

    return data


def unicode_to_string(input):
    """Convert all unicode strings to utf-8 strings
    """
    if isinstance(input, dict):
        return {
            unicode_to_string(key): unicode_to_string(
                value) for key, value in input.items()}
    elif isinstance(input, list):
        return [unicode_to_string(element) for element in input]
    elif isinstance(input, str):
        return input.encode('utf-8')
    else:
        return input


def corr_mat_filter(
    data: np.ndarray, stats: CorrStats, freqs: Tuple[float, float],
        order=3) -> np.ndarray:
    """ Filter a correlation matrix.

    Filters the correlation matrix corr_mat in the frequency band specified in
    freqs using a zero phase filter of twice the order given in order.

    :type data: np.ndarray
    :param data: correlation matrix
    :type stats: :class:`~seismic.correlate.stats.CorrStats`
    :param stats: The stats object corresponding to the CorrBulk object
    :type freqs: array-like of length 2
    :param freqs: lower and upper limits of the pass band in Hertz
    :type order: int
    :param order: half the order of the Butterworth filter

    :rtype: np.ndarray
    :return: filtered correlation matrix
    """

    if len(freqs) != 2:
        raise ValueError("freqs needs to be a two element array with the \
            lower and upper limits of the filter band in Hz.")

    # # end check

    fe = float(stats['sampling_rate']) / 2

    (b, a) = butter(
        order, np.array(freqs, dtype='float') / fe, btype='band')

    data = lfilter(b, a, data, axis=1)
    data = lfilter(b, a, data[:, ::-1], axis=1)[:, ::-1]

    return data


def corr_mat_resample(
    data: np.ndarray, stats: CorrStats, start_times: List[UTCDateTime],
        end_times=[]) -> Tuple[np.ndarray, CorrStats]:
    """ Function to create correlation matrices with constant sampling

    When created with recombine_corr_data the correlation matrix contains all
    available correlation streams but homogeneous sampling is not guaranteed
    as correlation functions may be missing due to data gaps. This function
    restructures the correlation matrix by inserting or averaging correlation
    functions to provide temporally homogeneous sampling. Inserted correlation
    functions consist of 'nan' if gaps are present and averaging is done if
    more than one correlation function falls in a bin between start_times[i]
    and end_times[i]. If end_time is an empty list (default) end_times[i] is
    set to start_times[i] + (start_times[1] - start_times[0])

    :type data: np.ndarray
    :param data: 2D matrix holding the correlation data
    :type Stats: :class:`~obspy.core.Stats`
    :param Stats: The stats object belonging to the
        :class:`~seismic.correlate.stream.CorrBulk` object.
    :type start_times: list of :class:`~obspy.UTCDateTime` objects
    :param start_times: list of starting times for the bins of the new
        sampling
    :type end_times: list of :class:`~obspy.UTCDateTime` objects
    :param end_times: list of starting times for the bins of the new sampling

    :rtype: tuple
    :return: the new data array and altered stats object.
    """

    if len(end_times) > 0 and (len(end_times) != len(start_times)):
        raise ValueError("end_times should be empty or of the same length as \
            start_times.")

    # old sampling times
    otime = np.array([ii.timestamp for ii in stats['corr_start']])
    if isinstance(start_times[0], UTCDateTime):
        start_times = [ii.timestamp for ii in start_times]

    # new sampling times
    stime = start_times
    if len(end_times):
        if isinstance(end_times[0], UTCDateTime):
            end_times = [ii.timestamp for ii in end_times]
        etime = end_times
    else:
        if len(start_times) == 1:
            # there is only one start_time given and no end_time => average all
            etime = np.array([stats['corr_end'][-1].timestamp])
        else:
            stime = np.array(stime)  # Cannot be a list for this operation
            etime = stime + (stime[1] - stime[0])

    # create masked array to avoid nans
    mm = np.ma.masked_array(
        data, np.isnan(data))

    # new corr_data matrix
    nmat = np.empty([len(stime), data.shape[1]])
    nmat.fill(np.nan)

    for ii in range(len(stime)):
        # index of measurements between start_time[ii] and end_time[ii]
        ind = np.nonzero((otime >= stime[ii]) * (otime < etime[ii]))  # ind is
        # a list(tuple) for dimensions
        if len(ind[0]) == 1:
            # one measurement found
            nmat[ii, :] = data[ind[0], :]
        elif len(ind[0]) > 1:
            # more than one measurement in range
            nmat[ii, :] = np.mean(mm[ind[0], :], 0).filled(np.nan)

    # assign new data
    data = nmat

    stats['corr_start'] = [UTCDateTime(st) for st in stime]
    stats['corr_end'] = [UTCDateTime(et) for et in etime]

    return data, stats


# Probably not necessary
def corr_mat_correct_decay(data: np.ndarray, stats: CorrStats) -> np.ndarray:
    """
    Correct for the amplitude decay in a correlation matrix.

    Due to attenuation and geometrical spreading the amplitude of the
    correlations decays with increasing lapse time. This decay is corrected
    by dividing the correlation functions by an exponential function that
    models the decay.

    :param data: Matrix holding correlation data
    :type data: np.ndarray
    :param stats: Stats object afiliated to a CorrBulk object
    :type stats: CorrStats
    :return: correlations with decay corrected amplitudes
    :rtype: np.ndarray
    """
    # copy input
    # cmat = deepcopy(corr_mat)

    # calculate the envelope of the correlation matrix
    env = corr_mat_envelope(data)

    # average causal and acausal part of the correlation matrix
    env, stats_mirr = corr_mat_mirror(env, stats)

    # average to a single correlation function
    env = np.nanmean(env, 0)  # Shouldn't that be a mean or divided by the dim?
    # env = np.nansum(env, 0)

    # fit the decay
    t = np.arange(stats_mirr['npts'])
    le = np.log(env)
    K, log_A = np.polyfit(t, le, 1)
    A = np.exp(log_A)

    # central sample
    cs = -stats['start_lag']*stats['sampling_rate']
    # time vector
    t = np.absolute(np.arange(stats['npts']) - cs)
    # theoretical envelope
    tenv = A * np.exp(K * t)

    # correct with theoretical envelope
    data /= np.tile(tenv, [data.shape[0], 1])

    return data


def corr_mat_envelope(data: np.ndarray) -> np.ndarray:
    """ Calculate the envelope of a correlation matrix.

    The corrlation data of the correlation matrix are replaced by their
    Hilbert envelopes.

    :type data: np.ndarray
    :param data: correlation matrix.

    :rtype: np.ndarray
    :return: **hilb**: The envelope of the correlation data in the same shape
        as the input array.
    """

    # calculate the Hilbert transform of the correlation data
    hilb = hilbert(data, axis=1)

    # replace corr_data with their envelopes
    hilb = np.absolute(hilb)

    return hilb


def corr_mat_normalize(
    data: np.ndarray, stats: CorrStats, starttime: float = None,
        endtime: float = None, normtype: str = 'energy') -> np.ndarray:
    """ Correct amplitude variations with time in a correlation matrix.

    Measure the maximum of the absolute value of the correlation matrix in
    a specified lapse time window and normalize the correlation traces by
    this values. A coherent phase in the respective lapse time window will
    have constant ampitude afterwards..

    :type data: np.ndarray
    :param data: correlation matrix from CorrBulk
    :type stats: :class:`~seismic.correlate.stats.CorrStats`
    :param stats: The stats object from the
        :class:`~seismic.correlate.stream.CorrBulk` object.
    :type starttime: float
    :param starttime: beginning of time window in seconds with respect to the
        zero position
    :type endtime: float
    :param endtime: end time window in seconds with respect to the zero
        position
    :type normtype: string
    :param normtype: one of the following 'energy', 'max', 'absmax', 'abssum'
        to decide about the way to calculate the normalization.

    :rtype: np.ndarray
    :return: Correlation matrix with normalized ampitudes.
    """

    # calculate indices of the time window
    # start
    # needs to be like this because this condition also has to work when it's 0
    if starttime is not None:
        # first sample
        start = starttime - stats['start_lag']
        start = int(np.floor(start*stats['sampling_rate']))
    else:
        start = None

    if endtime is not None:
        end = endtime - stats['start_lag']
        end = int(np.floor(end*stats['sampling_rate']))
    else:
        end = None

    # check range
    if start and start < 0:
        print('Error: starttime before beginning of trace. Data not changed.')
        return data
    if end and end >= stats['npts']:
        print('Error: endtime after end of trace. Data not changed.')
        return data

    # calculate normalization factors
    if normtype == 'energy':
        norm = np.sqrt(np.mean(data[:, start:end] ** 2, 1))
    elif normtype == 'abssum':
        norm = np.mean(np.abs(data[:, start:end]), 1)
    elif normtype == 'max':
        norm = np.max(data[:, start:end], axis=1)
    elif normtype == 'absmax':
        norm = np.max(np.abs(data[:, start:end]), axis=1)
    else:
        raise ValueError('Normtype is unknown.')

    # normalize the matrix
    norm[norm == 0] = 1
    data /= np.tile(np.atleast_2d(norm).T, (1, data.shape[1]))
    return data


def corr_mat_mirror(data: np.ndarray, stats: CorrStats) -> Tuple[
        np.ndarray, CorrStats]:
    """
    Average the causal and acausal parts of a correlation matrix.

    :param data: A matrix holding correlation data
    :type data: np.ndarray
    :param stats: the stats object corresponding to a CorrBulk object.
    :type stats: CorrStats
    :return: as the input but with
        avaraged causal and acausal parts of the correlation data
    :rtype: Tuple[np.ndarray, CorrStats]
    """

    zero_sample = -(stats['start_lag']*stats['sampling_rate'])

    if zero_sample <= 0:
        print('No data present for mirroring: starttime > zerotime.')
        return data, stats
    if stats['end_lag'] <= 0:
        print('No data present for mirroring: endtime < zerotime.')
        return data, stats
    if np.fmod(zero_sample, 1) != 0:
        print('need to shift for mirroring')
        return 0

    # estimate size of mirrored array
    acausal_samples = int(zero_sample + 1)
    causal_samples = int(stats['npts'] - acausal_samples + 1)
    # +1 because sample a zerotime counts twice
    size = np.max([acausal_samples, causal_samples])
    both = np.min([acausal_samples, causal_samples])

    # allocate array
    mir_data = np.zeros([data.shape[0], size])

    # fill the array
    mir_data[:, 0:causal_samples] = data[:, acausal_samples - 1:]
    mir_data[:, 0:acausal_samples] += data[:, acausal_samples - 1::-1]

    # divide by two where both are present
    mir_data[:, 0:both] /= 2.

    # adopt the stats
    mir_stats = stats.copy()
    mir_stats['start_lag'] = 0
    mir_stats['npts'] = size
    # -1 because of zero sample
    # mir_stats['end_lag'] = (size-1)/mir_stats['sampling_rate']

    return mir_data, mir_stats


def corr_mat_taper(
        data: np.ndarray, stats: CorrStats, width: float) -> np.ndarray:
    """ Taper a correlation matrix.

    Apply a taper to all traces in the correlation matrix.

    :type data: np.ndarray
    :param data: correlation matrix`
    :type width: float
    :param width: width to be tapered in seconds (per side)

    :rtype: np.ndarray
    :return: The tapered matrix

    ..note:: **In-place operation**
    """
    if width == 0:
        return data
    elif width > stats['npts']/stats['sampling_rate']:
        raise ValueError('Taper longer than signal.')
    width *= 2
    # calculate taper
    tap = cosine_taper(
        stats['npts'], width*stats['sampling_rate']/stats['npts'])

    # apply taper
    data *= np.tile(np.atleast_2d(tap), (data.shape[0], 1))

    return data


def corr_mat_taper_center(
    data: np.ndarray, stats: CorrStats, width: float,
        slope_frac: float = 0.05) -> np.ndarray:
    """
    Taper the central part of a correlation matrix.

    Due to electromagnetic cross-talk, signal processing or other effects the
    correlaton matrices are often contaminated around the zero lag time. This
    function tapers (multiples by zero) the central part of width `width`. To
    avoid problems with interpolation and filtering later on this is done with
    cosine taper

    :param data: Correlation data from CorrBulk object
    :type data: np.ndarray
    :param stats: Stats object from CorrBulk
    :type stats: CorrStats
    :param width: width of the central window to be tapered in seconds
        (in total, i.e. not per taper side)
    :type width: float
    :param slope_frac: fraction of `width` used for soothing of edges,
        defaults to 0.05
    :type slope_frac: float, optional
    :return: The tapered matrix
    :rtype: np.ndarray
    """
    if width == 0:
        return data
    elif width < 0:
        raise ValueError('Taper Width must not be negative.')
    elif width > stats['npts']/stats['sampling_rate']:
        raise ValueError('Taper width longer than signal.')

    # calculate size of taper (should be an even number)
    length = int(2. * np.ceil(width * stats['sampling_rate'] / 2.))
    slope_length = int(np.ceil(length * slope_frac / 2.))

    # calculate inverse taper
    taper = np.zeros(length + 1)
    tap = cosine_taper(slope_length * 2, p=1)
    taper[0:slope_length] = tap[slope_length:]
    taper[-slope_length:] = tap[:slope_length]

    # calculate start end end points of the taper
    start = int(-stats['start_lag']*stats['sampling_rate'] - length/2)
    if start < 0:
        taper = taper[np.absolute(start):]
        start = 0

    end = start + length
    if end > stats['npts']:
        end = stats['npts']
        taper = taper[0:end - start]
    # apply taper
    data[:, start:end + 1] *= \
        np.tile(np.atleast_2d(taper), (data.shape[0], 1))

    return data


def corr_mat_resample_time(
    data: np.ndarray, stats: CorrStats, freq: float) -> Tuple[
        np.ndarray, CorrStats]:
    """
    Resample the lapse time axis of a correlation matrix. The correlations are
    automatically filtered with a highpass filter of 0.4*sampling frequency to
    avoid aliasing.

    :type data: np.ndarray
    :param data: correlation matrix
    :type stats: :class:`~seismic.correlate.stats.CorrStats`
    :param stats: The stats object associated to the
        :class:`~seismic.correlate.stream.CorrBulk` object.
    :type freq: float
    :param freq: new sampling frequency

    :rtype: Tuple[np.ndarray, CorrStats]
    :return: Resampled data and changed stats object.

    ..note:: This action is performed **in-place**.
    """
    if freq > stats['sampling_rate']:
        raise ValueError('New sampling frequency higher than original f.')
    elif freq == stats['sampling_rate']:
        return data, stats

    # make last samples
    # x1 = n2*f1/f2-n1 + x2*f1/f2
    otime = float(stats['npts'])/stats['sampling_rate']
    num = int(np.floor(otime * freq))
    ntime = num/freq
    n1 = 0.
    n2 = 0.

    while ntime != otime:
        if ntime < otime:
            n1 += 1
            ntime = (num + n1)/freq
        else:
            n2 += 1
            otime = float(stats['npts']+n2)/stats['sampling_rate']
        if n1 > num:
            raise TypeError("No matching trace length found. Resampling not \
possible.")

    data = corr_mat_filter(data, stats, [0.001, 0.8*freq/2])

    data = np.concatenate(
        (data, np.zeros((data.shape[0], data.shape[1] + int(2*n2)))), axis=1)
    data = resample(data, int(2*(num + n1)), axis=1)[:, :num]
    stats['npts'] = data.shape[1]
    stats['sampling_rate'] = freq
    return data, stats


def corr_mat_decimate(
    data: np.ndarray, stats: CorrStats, factor: int) -> Tuple[
        np.ndarray, CorrStats]:
    """
    Downsample a correlation matrix by an integer sample. A low-pass filter
    is applied before decimating.

    :type data: np.ndarray
    :param data: correlation matrix
    :type stats: :class:`~seismic.correlate.stats.CorrStats`
    :param stats: The stats object associated to the
        :class:`~seismic.correlate.stream.CorrBulk` object.
    :type factor: int
    :param factor: The factor to reduce the sampling frequency by

    :rtype: Tuple[np.ndarray, CorrStats]
    :return: Decimated data and changed stats object.

    ..note:: This action is performed **in-place**.
    """
    if factor < 1:
        raise ValueError('Factor has to be larger than 1')
    elif factor == 1:
        return data, stats
    # apply low pass filter
    freq = stats['sampling_rate'] * 0.5 / float(factor)

    fe = float(stats['sampling_rate']) / 2

    (b, a) = butter(4, freq / fe, btype='lowpass')

    # forward
    data = lfilter(b, a, data, axis=1)
    # backwards
    data = lfilter(b, a, data[:, ::-1], axis=1)[:, ::-factor][:, :-1]
    # For some reason it includes the last sample
    stats['npts'] = data.shape[1]
    stats['sampling_rate'] = float(stats['sampling_rate'])/factor
    return data, stats


def corr_mat_resample_or_decimate(
    data: np.ndarray, stats: CorrStats, freq: float) -> Tuple[
        np.ndarray, CorrStats]:
    """
    Downsample a correlation matrix. Decides automatically whether to decimate
    or downsampled depneding on the target frequency. A low-pass filter
    is applied to avoid aliasing.

    :type data: np.ndarray
    :param data: correlation matrix
    :type stats: :class:`~seismic.correlate.stats.CorrStats`
    :param stats: The stats object associated to the
        :class:`~seismic.correlate.stream.CorrBulk` object.
    :type freq: float
    :param freq: new sampling frequency

    :rtype: Tuple[np.ndarray, CorrStats]
    :return: Decimated data and changed stats object.

    ..note:: This action is performed **in-place**.
    """
    sr = stats['sampling_rate']

    if sr/freq == sr//freq:
        return corr_mat_decimate(data, stats, int(sr//freq))
    else:
        return corr_mat_resample_time(data, stats, freq)


def corr_mat_extract_trace(
    data: np.ndarray, stats: CorrStats, method: str = 'mean',
        percentile: float = 50.) -> np.ndarray:
    """ Extract a representative trace from a correlation matrix.

    Extract a correlation trace from the that best represents the correlation
    matrix. ``Method`` decides about method to extract the trace. The following
    possibilities are available

    * ``mean`` averages all traces in the matrix
    * ``median`` extracts the median trace
    * ``norm_mean`` averages the traces normalized after normalizing for maxima
    * ``similarity_percentile`` averages the ``percentile`` % of traces that
        best correlate with the mean of all traces. This will exclude abnormal
        traces. ``percentile`` = 50 will return an average of traces with
        correlation (with mean trace) above the median.

    :type corr_mat: dictionary
    :param corr_mat: correlation matrix dictionary
    :type method: string
    :param method: method to extract the trace
    :type percentile: float
    :param percentile: only used for method=='similarity_percentile'

    :rtype: trace dictionary of type correlation trace
    :return **trace**: extracted trace
    """

    ndata = deepcopy(data)

    if method == 'mean':
        mm = np.ma.masked_array(
            ndata, np.isnan(ndata))
        out = np.mean(mm, 0).filled(np.nan)
    elif method == 'median':
        mm = np.ma.masked_array(
            ndata, np.isnan(ndata))
        out = np.median(mm, 0).filled(np.nan)
    elif method == 'norm_mean':
        # normalize the matrix
        ndata = corr_mat_normalize(ndata, stats, normtype='absmax')
        mm = np.ma.masked_array(
            ndata, np.isnan(ndata))
        out = (np.mean(mm, 0).filled(np.nan))
    elif method == 'similarity_percentile':
        # normalize the matrix
        ndata = corr_mat_normalize(ndata, stats, normtype='absmax')
        mm = np.ma.masked_array(
            ndata, np.isnan(ndata))
        # calc mean  trace
        mean_tr = np.mean(mm, 0).filled(np.nan)
        # calc similarity with mean trace
        tm_sq = np.sum(mean_tr ** 2)
        cm_sq = np.sum(ndata ** 2, axis=1)
        # cm_sq = np.sum(data ** 2, axis=1).filled(np.nan)
        cor = np.zeros(ndata.shape[0])
        for ind in range(mm.shape[0]):
            cor[ind] = (
                np.dot(ndata[ind, :], mean_tr) / np.sqrt(cm_sq[ind] * tm_sq))
        # estimate the percentile excluding nans
        tres = np.percentile(cor[~np.isnan(cor)], percentile)
        # find the traces that agree with the requested percentile and calc
        # their mean
        ind = cor > tres
        out = np.mean(ndata[ind, :], 0)
    else:
        raise ValueError("Method '%s' not defined." % method)

    return out


class Error(Exception):
    pass


class InputError(Error):
    def __init__(self,  msg):
        self.msg = msg


def corr_mat_stretch(
    cdata: np.ndarray, stats: CorrStats, ref_trc: np.ndarray = None,
    tw: List[np.ndarray] = None, stretch_range: float = 0.1,
    stretch_steps: int = 100, sides: str = 'both',
    return_sim_mat: bool = False,
    ref_tr_trim: Optional[Tuple[float, float]] = None,
        ref_tr_stats: Optional[CorrStats] = None) -> dict:
    """ Time stretch estimate through stretch and comparison.

    This function estimates stretching of the time axis of traces as it can
    occur if the propagation velocity changes.

    Time stretching is estimated comparing each correlation function stored
    in the ``corr_data`` matrix (one for each row) with ``stretch_steps``
    stretched versions  of reference trace stored in ``ref_trc``. If
    ``ref_trc`` is ``None`` the mean of all traces is used.
    The maximum amount of stretching may be passed in ``stretch_range``. The
    time axis is multiplied by 1/(1 + stretch).
    The best match (stretching amount and corresponding correlation value) is
    calculated on different time windows. If ``tw = None`` the stretching is
    estimated on the whole trace.

    :type corr_mat: dictionary
    :param corr_mat: correlation matrix dictionary as produced by
        :class:`~miic.core.macro.recombine_corr_data`
    :type ref_trc: :class:`~numpy.ndarray`
    :param ref_trc: 1D array containing the reference trace to be stretched
        and compared to the individual traces in ``mat``
    :type tw: list of :class:`~numpy.ndarray` of int
    :param tw: list of 1D ndarrays holding the indices of samples in the time
        windows to be use in the time shift estimate. The samples are counted
        from the zero lag time with the index of the first sample being 0. If
        ``tw = None`` the full time range is used.
    :type stretch_range: scalar
    :param stretch_range: Maximum amount of relative stretching.
        Stretching and compression is tested from ``-stretch_range`` to
        ``stretch_range``.
    :type stretch_steps: scalar`
    :param stretch_steps: Number of shifted version to be tested. The
        increment will be ``(2 * stretch_range) / stretch_steps``
    :type sides: str
    :param sides: Side of the reference matrix to be used for the stretching
        estimate ('both' | 'left' | 'right' | 'single') ``single`` is used for
        one-sided signals from active sources with zero lag time is on the
        first sample. Other options assume that the zero lag time is in the
        center of the traces.


    :rtype: Dictionary
    :return: **dv**: Dictionary with the following keys

        *corr*: 2d ndarray containing the correlation value for the best
            match for each row of ``mat`` and for each time window.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *value*: 2d ndarray containing the stretch amount corresponding to
            the best match for each row of ``mat`` and for each time window.
            Stretch is a relative value corresponding to the negative relative
            velocity change -dv/v.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *sim_mat*: 3d ndarray containing the similarity matricies that
            indicate the correlation coefficient with the reference for the
            different time windows, different times and different amount of
            stretching.
            Its dimension is: :py:func:`(len(tw),mat.shape[1],len(strvec))`
        *second_axis*: It contains the stretch vector used for the velocity
            change estimate.
        *vale_type*: It is equal to 'stretch' and specify the content of
            the returned 'value'.
        *method*: It is equal to 'single_ref' and specify in which "way" the
            values have been obtained.
    """

    data = deepcopy(cdata)

    dta = -stats['start_lag']
    dte = stats['end_lag']

    # format (trimm) the matrix for zero-time to be either at the beginning
    # or at the center as required by
    # miic.core.stretch_mod.time_stretch_estimate

    # In case a reference is provided but the matrix needs to be trimmed the
    # references also need to be trimmed. To do so we append the references to
    # the matrix, trimm it and remove the references again
    # if ref_trc is not None:
    #     rts = ref_trc.shape
    #     if len(rts) == 1:
    #         nr = 1
    #     else:
    #         nr = rts[0]
    #     data = np.concatenate((
    #         data, np.atleast_2d(ref_trc)), 0)
    #     reft = np.tile([UTCDateTime(1900, 1, 1)], (nr))
    #     stats['corr_start'] = np.concatenate((stats['corr_start'], reft), 0)

    # trim the matrices
    if sides == "single":
        # extract the time>0 part of the matrix
        data, stats = corr_mat_trim(data, stats, 0, dte)
        ref_tr_trim = (0, dte)
    else:
        # extract the central symmetric part (if dt<0 the trim will fail)
        dt = min(dta, dte)
        data, stats = corr_mat_trim(data, stats, -dt, dt)
        ref_tr_trim = (-dt, dt)

    # create or extract references
    if ref_trc is None:
        ref_trc = corr_mat_extract_trace(data, stats)
    # else:
    #     # extract and remove references from corr matrix again
    #     ref_trc = data[-nr:, :]
    #     data = data[:-nr, :]
    #     stats['corr_start'] = stats['corr_start'][:-nr]

    dv = multi_ref_vchange_and_align(
        data, ref_trc, tw=tw, stretch_range=stretch_range,
        stretch_steps=stretch_steps, sides=sides,
        return_sim_mat=return_sim_mat, ref_tr_trim=ref_tr_trim,
        ref_tr_stats=ref_tr_stats)

    # add the keys the can directly be transferred from the correlation matrix
    dv['stats'] = stats

    return dv


def corr_mat_shift(
    data: np.ndarray, stats: CorrStats, ref_trc: np.ndarray = None,
    tw: List[np.ndarray] = None, shift_range: float = 10,
    shift_steps: int = 101, sides: str = 'both',
        return_sim_mat: bool = False) -> dict:
    """ Time shift estimate through shifting and comparison.

    This function estimates shifting of the time axis of traces as it can
    occur if the clocks of digitizers drift.

    Time shifts are estimated comparing each correlation function stored
    in the ``corr_data`` matrix (one for each row) with ``shift_steps``
    shifted versions  of reference trace stored in ``ref_trc``.
    The maximum amount of shifting may be passed in ``shift_range``.
    The best match (shifting amount and corresponding correlation value) is
    calculated on different time windows. If ``tw = None`` the shifting is
    estimated on the whole trace.

    :type corr_mat: dictionary
    :param corr_mat: correlation matrix dictionary as produced by
        :class:`~miic.core.macro.recombine_corr_data`
    :type ref_trc: :class:`~numpy.ndarray`
    :param ref_trc: 1D array containing the reference trace to be stretched
        and compared to the individual traces in ``mat``
    :type tw: list of :class:`~numpy.ndarray` of int
    :param tw: list of 1D ndarrays holding the indices of sampels in the time
        windows to be use in the time shift estimate. The sampels are counted
        from the zero lag time with the index of the first sample being 0. If
        ``tw = None`` the full time range is used.
    :type shift_range: scalar
    :param shift_range: Maximum amount of shifting in samples.
        Shifting is tested from ``-stretch_range`` to
        ``stretch_range``.
    :type shift_steps: scalar`
    :param shift_steps: Number of shifted version to be tested. The
        increment will be ``(2 * shift_range) / shift_steps``
    :type sides: str
    :param sides: Side of the reference matrix to be used for the shifting
        estimate ('both' | 'right') ``single`` is used for
        one-sided signals from active sources with zero lag time is on the
        first sample. Other options assume that the zero lag time is in the
        center of the traces.


    :rtype: Dictionary
    :return: **dv**: Dictionary with the following keys

        *corr*: 2d ndarray containing the correlation value for the best
            match for each row of ``mat`` and for each time window.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *value*: 2d ndarray containing the stretch amount corresponding to
            the best match for each row of ``mat`` and for each time window.
            Stretch is a relative value corresponding to the negative relative
            velocity change -dv/v.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *sim_mat*: 3d ndarray containing the similarity matricies that
            indicate the correlation coefficient with the reference for the
            different time windows, different times and different amount of
            stretching.
            Its dimension is: :py:func:`(len(tw),mat.shape[1],len(strvec))`
        *second_axis*: It contains the stretch vector used for the velocity
            change estimate.
        *vale_type*: It is equal to 'shift' and specify the content of
            the returned 'value'.
        *method*: It is equal to 'single_ref' and specify in which "way" the
            values have been obtained.
    """

    data = deepcopy(data)

    dta = -stats.start_lag
    dte = stats.end_lag

    # format (trimm) the matrix for zero-time to be either at the beginning
    # or at the center as required by
    # miic.core.stretch_mod.time_stretch_estimate

    # In case a reference is provided but the matrix needs to be trimmed the
    # reference also needs to be trimmed. To do so we append the reference to
    # the matrix, trimm it and remove the reference again
    if ref_trc is not None:
        rts = ref_trc.shape
        if len(rts) == 1:
            nr = 1
        else:
            nr = rts[0]
        data = np.concatenate((
            data, np.atleast_2d(ref_trc)), 0)
        reft = np.tile([UTCDateTime(1900, 1, 1)], (nr))
        stats['corr_start'] = np.concatenate((stats['corr_start'], reft), 0)

    # trim the marices
    if sides == "single" or sides == 'right':
        # extract the time>0 part of the matrix
        data, stats = corr_mat_trim(data, stats, 0, dte)
    else:
        # extract the central symmetric part (if dt<0 the trim will fail)
        dt = min(dta, dte)
        data, stats = corr_mat_trim(data, stats, -dt, dt)

    # create or extract references
    if ref_trc is None:
        ref_trc = corr_mat_extract_trace(data, stats)
    else:
        # extract and remove references from corr matrix again
        ref_trc = np.squeeze(data[-nr:, :])
        data = data[:-nr, :]
        stats['corr_start'] = stats['corr_start'][:-nr]

    # set sides
    if sides == 'both':
        ss = False
    elif sides == 'right' or sides == 'single':
        ss = True
    else:
        raise ValueError(
            "Error: side is not recognized. Use either both or right.")

    dt = time_shift_estimate(
        data, ref_trc=ref_trc, tw=tw, shift_range=shift_range,
        shift_steps=shift_steps, single_sided=ss,
        return_sim_mat=return_sim_mat)

    # add the keys the can directly be transferred from the correlation matrix
    dt['stats'] = stats
    return dt


def measure_shift(
    data: np.ndarray, stats: CorrStats, ref_trc: Optional[np.ndarray] = None,
    tw: Optional[List[List]] = None, shift_range: float = 10,
    shift_steps: int = 101, sides: str = 'both',
        return_sim_mat: bool = False) -> List[DV]:
    """ Time shift estimate through shifting and comparison.

    This function estimates shifting of the time axis of traces as it can
    occur if the clocks of digitizers drift.

    Time shifts are estimated comparing each trace (e.g. correlation function
    stored in the ``corr_data`` matrix (one for each row) with shifted
    versions  of reference trace stored in ``ref_trc``. The range of shifting
    to be tested is given in ``shift_range`` in seconds. It is used in a
    symmetric way from -``shift_range`` to +``shift_range``. Shifting ist
    tested ``shift_steps`` times. ``shift_steps`` should be an odd number to
    test zero shifting. The best match (shifting amount and corresponding
    correlation value) is calculated in specified time windows. Multiple time
    windows may be specified in  ``tw``.


    :type data: :class:`~numpy.ndarray`
    :param data: correlation matrix or collection of traces. lag time runs
        along rows and number of traces is ``data.shape[0]``.
    :type ref_trc: :class:`~numpy.ndarray`
    :param ref_trc: 1D array containing the reference trace to be shifted
        and compared to the individual traces in ``data``. If ``ref_trc``is
        None, the average of all traces in ``data`` is used.
    :type tw: list of lists with two floats
    :param tw: list of lists that contain the start and end times of the time
        windows in which the similarity is to be estimated. Time are given in
        seconds of lag time with respect to zero lag time. Using multiple
        lists of start end end time the best shift can be estimated in
        multiple time windows. If ``tw = None`` one time window is used
        containing the full time range of lag times in the traces.
    :type shift_range: scalar
    :param shift_range: Maximum amount of shifting in second.
        Shifting is tested from ``-stretch_range`` to
        ``stretch_range``.
    :type shift_steps: scalar`
    :param shift_steps: Number of shifted version to be tested. The
        increment will be ``(2 * shift_range) / shift_steps``
    :type sides: str
    :param sides: Side of the traces to be used for the shifting estimate
        ('both' | 'single'). ``single`` is used for
        one-sided signals from active sources or if the time window shall
        not be symmetric. For ``both`` the time window will be mirrowd about
        zero lag time, e.g. [start,end] will result in time windows
        [-end:-start] and [start:end] being used simultaneousy


    :rtype: List[Dictionary]
    :return: **dv**: List of ``len(tw)`` Dictionaries with the following keys

        *corr*: 2d ndarray containing the correlation value for the best
            match for each row of ``mat`` and for each time window.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *value*: 2d ndarray containing the stretch amount corresponding to
            the best match for each row of ``mat`` and for each time window.
            Stretch is a relative value corresponding to the negative relative
            velocity change -dv/v.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *sim_mat*: 3d ndarray containing the similarity matricies that
            indicate the correlation coefficient with the reference for the
            different time windows, different times and different amount of
            stretching.
            Its dimension is: :py:func:`(len(tw),mat.shape[1],len(strvec))`
        *second_axis*: It contains the stretch vector used for the velocity
            change estimate.
        *vale_type*: It is equal to 'shift' and specify the content of
            the returned 'value'.
        *method*: It is equal to 'single_ref' and specify in which "way" the
            values have been obtained.
    """

    if sides not in ['both', 'single']:
        raise ValueError(
            f"'sides' must be either 'both' or 'single', not {sides}")
    if tw is not None:
        for twi in tw:
            if not len(twi) == 2:
                raise ValueError(
                    f"individual time windows must be specified as two-item"
                    f"list, not {twi} of length {len(twi)}.")

        if sides == 'both':
            if np.any(np.array(tw) < 0):
                raise ValueError(
                    "For 'sides=='both' all values in tw must be >= 0")
    if ref_trc is not None:
        if len(ref_trc) != data.shape[1]:
            raise ValueError(
                "Length of 'ref_trc' must match 'data.shape[1]', "
                f"they are {len(ref_trc)} and {data.shape[1]}.")

    data = deepcopy(data)
    stats = deepcopy(stats)

    # If reference trace is given attach it to the data for common
    # preprocessing
    if ref_trc is not None:
        data = np.concatenate((
            data, np.atleast_2d(ref_trc)), 0)
        reft = np.tile([UTCDateTime(1900, 1, 1)], (1))
        stats['corr_start'] = np.concatenate((stats['corr_start'], reft), 0)
    # If no time window is given use the whole range - shift_range
    if tw is None:
        tw = [[stats.start_lag + shift_range, stats.end_lag - shift_range]]

    # trim to required lapse time range
    twmax = np.max(np.array(tw)[:]) + shift_range
    if sides == 'both':
        twmin = -twmax
    else:
        twmin = np.min(np.array(tw)[:]) - shift_range
    assert twmax <= stats.end_lag, "Time window ends later than the trace. " \
        f"Latest time window + shift_range ends at {twmax}s " \
        f"while trace end at {stats.end_lag}s."
    assert twmin >= stats.start_lag, f"Time window starts earlier than the " \
        f"trace. Earliest time window - shift_range starts at {twmin}s " \
        f"while trace starts at {stats.start_lag}s."
    data, stats = corr_mat_trim(data, stats, twmin, twmax)

    # create or extract references
    if ref_trc is None:
        ref_trc = corr_mat_extract_trace(data, stats)
    else:
        # extract and remove references from corr matrix again
        ref_trc = np.squeeze(data[-1:, :])
        data = data[:-1, :]
        stats['corr_start'] = stats['corr_start'][:-1]

    # create matrix with shifted references
    shifts = np.linspace(-shift_range, shift_range, shift_steps)
    ref_mat = create_shifted_ref_mat(ref_trc, stats, shifts)

    # create space for results
    dt_list = []
    # do computation for each time window
    for twi in tw:
        # create indices of time windows
        indices = np.arange(
            np.ceil(twi[0]*stats.sampling_rate),
            np.floor(twi[1]*stats.sampling_rate))
        if sides == 'both':
            indices = np.concatenate((np.flipud(-indices), indices))
        indices -= np.round(stats.start_lag * stats.sampling_rate)
        indices = indices.astype(int)
        # compare data and shifted reference
        sim_mat = compare_with_modified_reference(data, ref_mat, indices)
        corr = sim_mat.max(axis=1)
        value = shifts[sim_mat.argmax(axis=1)]
        # Set dt to NaN where the correlation is NaN instead of having it equal
        # to one of the two stretch_range limits
        value[np.isnan(corr)] = np.nan
        # assemble results
        dt = {
            'stats': stats,
            'corr': np.squeeze(corr),
            'value': np.squeeze(value),
            'second_axis': shifts,
            'value_type': np.array(['shift']),
            'method': np.array(['absolute_shift'])}
        if return_sim_mat:
            dt.update({'sim_mat': np.squeeze(sim_mat)})
        else:
            dt.update({'sim_mat': None})

        dt_list.append(DV(**dt))

    return dt_list


def apply_shift(
    data: np.ndarray, stats: CorrStats, shifts: np.ndarray) -> Tuple[
        np.ndarray, CorrStats]:
    """ Apply a shift to a correlation matrix.

    This function applies a shift given in seconds to each trace in the matrix.

    :type data: :class:`~numpy.ndarray`
    :param data: correlation matrix or collection of traces. lag time runs
        along rows and number of traces is ``data.shape[0]``.
    :type shifts: :class:`~numpy.ndarray`
    :param shifts: shifts in seconds for each trace in 'data'
    """
    times = np.linspace(stats.start_lag, stats.end_lag, stats.npts)
    # stretch every line
    for (ii, line) in enumerate(data):
        # s = UnivariateSpline(times, line, s=0)
        f = interp1d(
            times, line, fill_value='extrapolate')
        data[ii, :] = f(times+shifts[ii])

    return data


def apply_stretch(
    data: np.ndarray, stats: CorrStats, stretches: np.ndarray) -> Tuple[
        np.ndarray, CorrStats]:
    """ Apply a stretches to a correlation matrix.

    This function applies a stretch given in relative units to each trace
    in the matrix.

    :type data: :class:`~numpy.ndarray`
    :param data: correlation matrix or collection of traces. lag time runs
        along rows and number of traces is ``data.shape[0]``.
    :type stretches: :class:`~numpy.ndarray`
    :param stretches: stretches in relative units for each trace in 'data'
    """
    times = np.linspace(stats.start_lag, stats.end_lag, stats.npts)

    # stretch every line
    for (ii, line) in enumerate(data):
        # Extrapolating stretches can lead to weird boundary values
        # So we set the boundary to 0 instead of extrapolating (ext=1)
        s = UnivariateSpline(times, line, s=0, ext=1)
        # dv defined as difference
        # data[ii, :] = s(times * (1 + stretches[ii]))
        # dv defined as logarithmic change
        data[ii, :] = s(times * np.exp(-stretches[ii]))
    return data, stats



'''
:copyright:
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Tuesday, 15th June 2021 03:42:14 pm
Last Modified: Monday, 7th August 2023 08:15:39 pm
'''
from typing import List, Tuple, Optional
from copy import deepcopy

import numpy as np
from obspy.signal.invsim import cosine_taper
from scipy.interpolate import UnivariateSpline, interp1d
from seismic.correlate.stats import CorrStats
from seismic.monitor.trim import corr_mat_trim


def time_windows_creation(
        starting_list: list, t_width: List[int] or int) -> np.ndarray:
    """ Time windows creation.

    A matrix containing one time window for each row is created. The starting
    samples of each one of them are passed in the ``starting_list`` parameter.
    The windows length ``t_width`` can be scalar or a list of values.
    In the latter case both lists ``starting_list`` and ``t_width`` must
    have the same length.

    :type starting_list: list
    :param starting_list: List of starting points
    :type t_width: int or list of int
    :param t_width: Windows length

    :rtype: :class:`~numpy.ndarray`
    :return: **tw_mat**: 2d ndarray containing the indexes of one time window
        for each row
    """

    if not np.isscalar(starting_list):
        if not np.isscalar(t_width) and len(t_width) != len(starting_list):
            raise ValueError("t_width must be a scalar or list of scalars of\
                            the same length as starting_list")
    if np.any(np.array(t_width) <= 0):
        raise ValueError('t_width must only contain values greater than 0.')

    tw_list = []

    if np.isscalar(starting_list):
        if np.isscalar(t_width):
            wlen = t_width
        else:
            wlen = t_width[0]
        tw_list.append(np.arange(starting_list, starting_list + wlen, 1))
    else:
        for (ii, cstart) in enumerate(starting_list):
            if np.isscalar(t_width):
                wlen = t_width
            else:
                wlen = t_width[ii]
            tw_list.append(np.arange(cstart, cstart + wlen, 1))

    return tw_list


def compute_wfc(
    mat: np.ndarray, tw: np.ndarray, refcorr: np.ndarray, sides: str,
        remove_nans: bool = True) -> np.ndarray:
    """
    Computes the waveform coherency (**WFC**) between the given reference
    correlation and correlationmatrix.

    See Steinmann, et. al. (2021) for details

    :param mat: 2D Correlation Matrix (function of corrstart and time lag)
    :type mat: np.ndarray
    :param tw: Lag time window to use for the computation
    :type tw: np.ndarray
    :param refcorr: 1D reference Correlation Trace extracted from `mat`.
    :type refcorr: np.ndarray
    :param sides: Which sides to use. Can be `both`, `right`, `left`,
        or `single`.
    :type sides: str
    :param remove_nans: Remove nans from CorrMatrix, defaults to True
    :type remove_nans: bool, optional
    :raises ValueError: unknown option in sides
    :return: A 1D array with len=N(time windows)
    :rtype: np.ndarray
    """
    # Mat must be a 2d vector in every case so
    mat = np.atleast_2d(mat)

    assert len(refcorr) == mat.shape[1]

    if remove_nans:
        mat = np.nan_to_num(mat)

    center_p = np.floor((mat.shape[1] - 1.) / 2.)

    corr = np.zeros((len(tw), mat.shape[0]))

    for (ii, ctw) in enumerate(tw):

        if sides == 'both':
            if ctw[0] == 0:
                ctw = np.hstack((
                    center_p - ctw[::-1],
                    center_p + ctw[1:])).astype(np.int32)
            else:
                ctw = np.hstack((
                    center_p - ctw[::-1],
                    center_p + ctw)).astype(np.int32)
        elif sides == 'left':
            ctw = (center_p - ctw[::-1]).astype(np.int32)
        elif sides == 'right':
            ctw = (center_p + ctw).astype(np.int32)
        elif sides == 'single':
            ctw = ctw.astype(np.int32)
        else:
            raise ValueError(
                'sides = %s not a valid option.' % sides)

        mask = np.zeros((mat.shape[1],))
        mask[ctw] = 1

        ref_mask_mat = mask
        mat_mask_mat = np.tile(mask, (mat.shape[0], 1))

        first = mat * mat_mask_mat
        second = refcorr * ref_mask_mat

        dprod = np.dot(first, second.T)

        # Normalization
        f_sq = np.sum(first ** 2, axis=1)
        s_sq = np.sum(second ** 2)

        f_sq = f_sq.reshape(1, len(f_sq))

        den = np.sqrt(np.dot(f_sq, s_sq))

        corr[ii] = dprod/den
    return corr


def wfc_multi_reftr(
    corr_data: np.ndarray, ref_trs: np.ndarray, tw: np.ndarray, sides: str,
        remove_nans: bool = True) -> dict:
    """
    Computes the waveform coherency (**WFC**) between the given reference
    correlation(s) and correlation matrix.
    If several references are given, it will loop over these

    See Steinmann, et. al. (2021) for details

    :param corr_data: 2D Correlation Matrix
        (function of corrstart and time lag)
    :type corr_data: np.ndarray
    :param refcorr: 1 or 2D reference Correlation Trace extracted from
        `corr_data`.
        If 2D, it will be interpreted as one refcorr trace per row.
    :type refcorr: np.ndarray
    :param tw: Lag time window to use for the computation
    :type tw: np.ndarray
    :param sides: Which sides to use. Can be `both`, `right`, `left`,
        or `single`.
    :type sides: str
    :param remove_nans: Remove nans from CorrMatrix, defaults to True
    :return: A dictionary with one correlation array for each reference trace.
    :rtype: dict
    """
    if remove_nans:
        corr_data = np.nan_to_num(corr_data)
        ref_trs = np.nan_to_num(ref_trs)

    # remove 1-dimensions
    ref_trs = np.squeeze(ref_trs)

    multi_ref_panel = {}

    # check how many reference traces have been passed
    try:
        reftr_count, _ = ref_trs.shape
    except ValueError:  # An array is passed
        reftr_count = 1

    if reftr_count == 1:
        key = "reftr_0"
        value = compute_wfc(
            corr_data, tw, ref_trs, sides, remove_nans=remove_nans)
        multi_ref_panel.update({key: value})
    else:  # For multiple-traces loops
        for i, rftr in enumerate(ref_trs):
            key = "reftr_%d" % i
            value = value = compute_wfc(
                corr_data, tw, rftr, sides, remove_nans=remove_nans)
            multi_ref_panel.update({key: value})
    return multi_ref_panel


def velocity_change_estimate(
    mat: np.ndarray, tw: np.ndarray, strrefmat: np.ndarray, strvec: np.ndarray,
    sides: str = 'both', return_sim_mat: bool = False,
        remove_nans: bool = True) -> dict:
    """ Velocity change estimate through stretching and comparison.

    Velocity changes are estimated comparing each correlation function stored
    in the ``mat`` matrix (one for each row) with ``strrefmat.shape[0]``
    stretched versions  of a reference trace stored in ``strrefmat``.
    The stretch amount for each row of ``strrefmat`` must be passed in the
    ``strvec`` vector.
    The best match (stretch amount and corresponding correlation value) is
    calculated on different time windows (each row of ``tw`` is a different
    one) instead of on the whole trace.

    :type mat: :class:`~numpy.ndarray`
    :param mat: 2d ndarray containing the correlation functions.
        One for each row.
    :type tw: :class:`~numpy.ndarray` of int
    :param tw: 2d ndarray of time windows to be use in the velocity change
         estimate.
    :type strrefmat: :class:`~numpy.ndarray`
    :param strrefmat: 2D array containing stretched version of the reference
         matrix
    :type strvec: :class:`~numpy.ndarray` or list
    :param strvec: Stretch amount for each row of ``strrefmat``
    :type sides: string
    :param sides: Side of the reference matrix to be used for the velocity
        change estimate ('both' | 'left' | 'right' | 'single')
    :type remove_nans: bool
    :param remove_nans: If `True` applay :func:`~numpy.nan_to_num` to the
        given correlation matrix before any other operation.

    :rtype: Dictionary
    :return: **dv**: Dictionary with the following keys

        *corr*: 2d ndarray containing the correlation value for the best
            match for each row of ``mat`` and for each time window.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *value*: 2d ndarray containing the stretch amount corresponding to
            the best match for each row of ``mat`` and for each time window.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *sim_mat*: 3d ndarray containing the similarity matricies that
            indicate the correlation coefficient with the reference for the
            different time windows, different times and different amount of
            stretching.
            Its dimension is: :py:func:`(len(tw),mat.shape[1],len(strvec))`
        *second_axis*: It contains the stretch vector used for the velocity
            change estimate.
        *vale_type*: It is equal to 'stretch' and specify the content of
            the returned 'value'.
        *method*: It is equal to 'single_ref' and specify in which "way" the
            values have been obtained.
    """

    # Mat must be a 2d vector in every case so
    mat = np.atleast_2d(mat)

    if strrefmat.shape[1] != mat.shape[1]:
        raise ValueError(
            'STRREFMAT and MAT must have the same number of'
            f"samples corrmat has shape {mat.shape} "
            f'and stretched reference matrix has {strrefmat.shape}')

    if remove_nans:
        mat = np.nan_to_num(mat)

    center_p = np.floor((mat.shape[1] - 1.) / 2.)

    nstr = strrefmat.shape[0]

    corr = np.zeros((len(tw), mat.shape[0]))
    dt = np.zeros((len(tw), mat.shape[0]))
    sim_mat = np.zeros([mat.shape[0], len(strvec), len(tw)])

    for (ii, ctw) in enumerate(tw):

        if sides == 'both':
            # ctw = np.hstack((center_p - ctw[::-1],
            #                  center_p + ctw)).astype(np.int32)
            if ctw[0] == 0:
                ctw = np.hstack((
                    center_p - ctw[::-1],
                    center_p + ctw[1:])).astype(np.int32)
            else:
                ctw = np.hstack((
                    center_p - ctw[::-1],
                    center_p + ctw)).astype(np.int32)
        elif sides == 'left':
            ctw = (center_p - ctw[::-1]).astype(np.int32)
        elif sides == 'right':
            ctw = (center_p + ctw).astype(np.int32)
        elif sides == 'single':
            ctw = ctw.astype(np.int32)
        else:
            raise ValueError(
                'sides = %s not a valid option.' % sides)

        mask = np.zeros((mat.shape[1],))
        mask[ctw] = 1

        ref_mask_mat = np.tile(mask, (nstr, 1))
        mat_mask_mat = np.tile(mask, (mat.shape[0], 1))

        first = mat * mat_mask_mat
        second = strrefmat * ref_mask_mat

        dprod = np.dot(first, second.T)

        # Normalization
        f_sq = np.sum(first ** 2, axis=1)
        s_sq = np.sum(second ** 2, axis=1)

        f_sq = f_sq.reshape(1, len(f_sq))
        s_sq = s_sq.reshape(1, len(s_sq))

        den = np.sqrt(np.dot(f_sq.T, s_sq))

        tmp = dprod / den
        sim_mat[:, :, ii] = tmp

        tmp_corr_vect = tmp.max(axis=1)
        corr[ii, :] = tmp_corr_vect
        dt[ii, :] = strvec[tmp.argmax(axis=1)]

        # Set dt to NaN where the correlation is NaN instead of having it equal
        # to one of the two stretch_range limits
        dt[ii, np.isnan(tmp_corr_vect)] = np.nan

    dv = {'corr': np.squeeze(corr),
          'value': np.squeeze(dt),
          'second_axis': strvec,
          'value_type': 'stretch',
          'method': 'single_ref'}

    if return_sim_mat:
        dv.update({'sim_mat': np.squeeze(sim_mat)})

    return dv


def time_stretch_estimate(
    corr_data: np.ndarray, ref_trc: np.ndarray = None, tw: np.ndarray = None,
    stretch_range: float = 0.1, stretch_steps: int = 100, sides: str = 'both',
    remove_nans: bool = True,
    ref_tr_trim: Optional[Tuple[float, float]] = None,
        ref_tr_stats=None) -> dict:
    """ Time stretch estimate through stretch and comparison.

    This function estimates stretching of the time axis of traces as it can
    occur if the propagation velocity changes.

    Time stretching is estimated comparing each correlation function stored
    in the ``corr_data`` matrix (one for each row) with ``stretch_steps``
    stretched versions  of reference trace stored in ``ref_trc``.
    The maximum amount of stretching may be passed in ``stretch_range``. The
    time axis is multiplied by 1/(1 + stretch).
    The best match (stretching amount and corresponding correlation value) is
    calculated on different time windows. If ``tw = None`` the stretching is
    estimated on the whole trace.

    :type corr_data: :class:`~numpy.ndarray`
    :param corr_data: 2d ndarray containing the correlation functions.
        One for each row.
    :type ref_trc: :class:`~numpy.ndarray`
    :param ref_trc: 1D array containing the reference trace to be shifted and
        compared to the individual traces in ``mat``
    :type tw: list of :class:`~numpy.ndarray` of int
    :param tw: list of 1D ndarrays holding the indices of sampels in the time
        windows to be use in the time shift estimate. The sampels are counted
        from the zero lag time with the index of the first sample being 0. If
        ``tw = None`` the full time range is used.
    :type stretch_range: scalar
    :param stretch_range: Maximum amount of relative stretching.
        Stretching and compression is tested from ``-stretch_range`` to
        ``stretch_range``.
    :type stretch_steps: scalar`
    :param stretch_steps: Number of shifted version to be tested. The
        increment will be ``(2 * stretch_range) / stretch_steps``
    :type sides: str
    :param sides: Side of the reference matrix to be used for the stretching
        estimate ('both' | 'left' | 'right' | 'single') ``single`` is used for
        one-sided signals from active sources with zero lag time is on the
        first sample. Other options assume that the zero lag time is in the
        center of the traces.
    :type remove_nans: bool
    :param remove_nans: If `True` applay :func:`~numpy.nan_to_num` to the
        given correlation matrix before any other operation.

    :rtype: Dictionary
    :return: **dv**: Dictionary with the following keys

        *corr*: 2d ndarray containing the correlation value for the best
            match for each row of ``mat`` and for each time window.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *value*: 2d ndarray containing the stretch amount corresponding to
            the best match for each row of ``mat`` and for each time window.
            Stretch is a relative value corresponding to the negative relative
            velocity change -dv/v.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *sim_mat*: 3d ndarray containing the similarity matricies that
            indicate the correlation coefficient with the reference for the
            different time windows, different times and different amount of
            stretching.
            Its dimension is: :py:func:`(len(tw),mat.shape[1],len(strvec))`
        *second_axis*: It contains the stretch vector used for the velocity
            change estimate.
        *vale_type*: It is equal to 'stretch' and specify the content of
            the returned 'value'.
        *method*: It is equal to 'single_ref' and specify in which "way" the
            values have been obtained.
    """

    # Mat must be a 2d vector in every case so
    mat = np.atleast_2d(corr_data)

    # FIX: Trace back this problem to the original source and remove
    # this statement
    if remove_nans:
        mat = np.nan_to_num(mat)

    # generate the reference trace if not given (use the whole time span)
    if ref_trc is None:
        ref_trc = np.nanmean(mat, axis=0)

    # generate time window if not given (use the full length of the correlation
    # trace)
    if tw is None:
        if sides == 'single':
            tw = time_windows_creation([0], [mat.shape[1]])
        else:
            tw = time_windows_creation([0], [int(np.floor(mat.shape[1] / 2.))])

    # taper and extend the reference trace to avoid interpolation
    # artefacts at the ends of the trace
    # 01.07.21
    # This taper decreases the correlation between the traces,
    # just don't permit extrapolation
    # taper = cosine_taper(len(ref_trc), 0.05)
    # ref_trc *= taper

    # different values of shifting to be tested
    stretches = np.linspace(-stretch_range, stretch_range, stretch_steps)
    # dv defined as difference
    # time_facs = 1/(1 + stretchs)
    # dv defined as logarithmic stretch
    time_facs = np.exp(-stretches)

    # time axis
    if sides != 'single':
        time_idx = np.arange(len(ref_trc)) - (len(ref_trc) - 1.) / 2.
    else:
        time_idx = np.arange(len(ref_trc))

    # create the array to hold the shifted traces
    ref_stretch = np.zeros((len(stretches), len(ref_trc)))

    # create a spline object for the reference trace
    # :NOTE: This change can give very different results and should
    # **definitely** be discussed!
    # 01.07.21 No extrapolation
    # 15.11.22 set extrapolation to 0 / ext=1
    ref_tr_spline = UnivariateSpline(time_idx, ref_trc, s=0, ext='zeros')

    # evaluate the spline object at different points and put in the prepared
    # array
    for (k, this_fac) in enumerate(time_facs):
        ref_stretch[k, :] = ref_tr_spline(time_idx * this_fac)

    if ref_tr_trim is not None:
        ref_stretch, _ = corr_mat_trim(
            ref_stretch, deepcopy(ref_tr_stats), ref_tr_trim[0],
            ref_tr_trim[1])

    # search best fit of the crosscorrs to one of the stretched ref_traces
    dv = velocity_change_estimate(
        mat, tw, ref_stretch, stretches, sides=sides, return_sim_mat=True,
        remove_nans=remove_nans)

    # TODO: It is not really clear why it it necessary to transpose here so
    # this is the fist point where to look in case of errors.
    dv['corr'] = dv['corr'].T
    dv['value'] = dv['value'].T
    # dv.update({'stretch_vec': stretchs})

    return dv


def multi_ref_vchange(
    corr_data: np.ndarray, ref_trs: np.ndarray, tw: np.ndarray = None,
    stretch_range: float = 0.1, stretch_steps: int = 100, sides: str = 'both',
    remove_nans: bool = True,
    ref_tr_trim: Optional[Tuple[float, float]] = None,
        ref_tr_stats=None) -> dict:
    """ Velocity change estimate with single or multiple reference traces.

    This function estimates the velocity change corresponding to each row of
    the ``corr_data`` matrix respect to the reference trace/s passed in
    ``ref_trs``.

    The velocity change is estimated comparing each correlation function stored
    in the ``corr_data`` matrix (one for each row) with ``stretch_steps``
    stretched versions of reference/s trace stored in ``ref_trs``.
    The maximum amount of stretching may be passed in ``stretch_range``.
    The best match (stretching amount and corresponding correlation value) is
    calculated on different time windows. If ``tw = None`` the stretching is
    estimated on the whole trace.

    The output is a dictionary with keys of the form ``"reftr_%d" % i``: One
    for each reference trace. The corresponding ``value`` is aslo a dictionary
    that has a structure conforming with
    :py:class:`~miic.core.stretch_mod.time_stretch_estimate` output.

    :type corr_data: :class:`~numpy.ndarray`
    :param corr_data: 2d ndarray containing the correlation functions.
        One for each row.
    :type ref_trc: :class:`~numpy.ndarray`
    :param ref_trc: 1D array containing the reference trace to be shifted and
        compared to the individual traces in ``mat``
    :type tw: list of :class:`~numpy.ndarray` of int
    :param tw: list of 1D ndarrays holding the indices of sampels in the time
        windows to be use in the time shift estimate. The sampels are counted
        from the zero lag time with the index of the first sample being 0. If
        ``tw = None`` the full time range is used.
    :type stretch_range: scalar
    :param stretch_range: Maximum amount of relative stretching.
        Stretching and compression is tested from ``-stretch_range`` to
        ``stretch_range``.
    :type stretch_steps: scalar`
    :param stretch_steps: Number of shifted version to be tested. The
        increment will be ``(2 * stretch_range) / stretch_steps``
    :type single_sided: bool
    :param single_sided: if True zero lag time is on the first sample. If
        False the zero lag time is in the center of the traces.
    :type sides: str
    :param sides: Side of the reference matrix to be used for the stretching
        estimate ('both' | 'left' | 'right' | 'single') ``single`` is used for
        one-sided signals from active sources with zero lag time is on the
        first sample. Other options assume that the zero lag time is in the
        center of the traces.
    :type remove_nans: bool
    :param remove_nans: If `True` applay :func:`~numpy.nan_to_num` to the
        given correlation matrix before any other operation.

    :rtype: dictionary
    :return: **multi_ref_panel**: It is a dictionary that contains as much
        dictionaries as reference traces have been used. The key format is
        ``"reftr_%d" % i`` and the corresponding value is also a dictionary
        with the structure described in
        :py:class:`~miic.core.stretch_mod.time_stretch_estimate`
    """

    # FIX: Trace back this problem to the original source and remove
    # this statement
    if remove_nans:
        corr_data = np.nan_to_num(corr_data)
        ref_trs = np.nan_to_num(ref_trs)

    # remove 1-dimensions
    ref_trs = np.squeeze(ref_trs)

    # check how many reference traces have been passed
    try:
        reftr_count, _ = ref_trs.shape
    except ValueError:  # An array is passed
        reftr_count = 1

    # Dictionary that will hold all the results
    multi_ref_panel = {}

    # When there is just 1 reference trace no loop is necessary
    if reftr_count == 1:
        key = "reftr_0"
        value = time_stretch_estimate(
            corr_data, ref_trc=ref_trs, tw=tw, stretch_range=stretch_range,
            stretch_steps=stretch_steps, sides=sides, remove_nans=remove_nans,
            ref_tr_trim=ref_tr_trim, ref_tr_stats=ref_tr_stats)
        multi_ref_panel.update({key: value})
    else:  # For multiple-traces loops
        for i in range(reftr_count):
            ref_trc = ref_trs[i]
            key = "reftr_%d" % int(i)
            value = time_stretch_estimate(
                corr_data, ref_trc=ref_trc, tw=tw, stretch_range=stretch_range,
                stretch_steps=stretch_steps, sides=sides,
                remove_nans=remove_nans,
                ref_tr_trim=ref_tr_trim, ref_tr_stats=ref_tr_stats)
            multi_ref_panel.update({key: value})
    return multi_ref_panel


def est_shift_from_dt_corr(
    dt1: np.ndarray, dt2: np.ndarray, corr1: np.ndarray,
        corr2: np.ndarray) -> Tuple[float, float]:
    """ Estimation of a baseline shift between velocity-change measurements
    preformed with different references.

    The use of different reference traces obtaind from different reference
    periods will result in a shift of the velocity-change curves that ideally
    characterizes the velocity variation between the two reference periods.
    Instead of directly measuring this velocity change from the two reference
    traces it is calulated here as the weighted average of the point
    differences between the two velocity-change curves weighted by their
    inverse variance according to Weaver et al. GJI 2011 (On the precision of
    noise correlation interferometry)

    Input vertors must all be of the same lenth.

    :type dt1: :class:`~numpy.ndarray`
    :pram dt1: Velocity variation measured for reference A
    :type dt2: :class:`~numpy.ndarray`
    :pram dt2: Velocity variation measured for reference B
    :type corr1: :class:`~numpy.ndarray`
    :pram corr1: Correlation between velocity corrected trace and reference A
    :type corr2: :class:`~numpy.ndarray`
    :pram corr2: Correlation between velocity corrected trace and reference B
    """

    # Create masked arrays so that NaNs are handled properly
    dt1 = np.ma.masked_array(
        dt1, mask=(np.isnan(dt1) | np.isinf(dt1)), fill_value=0)
    dt2 = np.ma.masked_array(
        dt2, mask=(np.isnan(dt2) | np.isinf(dt2)), fill_value=0)

    corr1 = np.ma.masked_array(
        corr1, mask=(np.isnan(corr1) | np.isinf(corr1)), fill_value=0)
    corr2 = np.ma.masked_array(
        corr2, mask=(np.isnan(corr2) | np.isinf(corr2)), fill_value=0)

    # October 21 - PM - Not entirely sure, why this is done?
    # However, it leads to problems when creating the mask
    # (because shape wrong). Two options: 1.DOn't do it or
    # 2. Also remove the correspdoning points in the dt objects
    # Remove the points where the correlation is 0
    no_zero = (corr1 > 0) & (corr2 > 0)
    # corr1 = corr1[no_zero]
    # corr2 = corr2[no_zero]

    # Estimate the point-variance for the two curves
    # var1 = (1 - corr1[no_zero] ** 2) / (4 * corr1[no_zero] ** 2)
    # var2 = (1 - corr2[no_zero] ** 2) / (4 * corr2[no_zero] ** 2)
    var1 = (1 - corr1 ** 2) / (4 * corr1 ** 2)
    var2 = (1 - corr2 ** 2) / (4 * corr2 ** 2)

    # Calculate the point-weight
    wgt = 1 / (var1 + var2)
    # Insert he no_zero mask here instead.
    mask = (corr1 > 0.999) & (corr2 > 0.999) & (~no_zero)
    wgt = wgt[~mask]

    # This saves from returning a masked value
    try:
        wgt_s = np.sum(wgt).filled(np.nan)
    except Exception:
        wgt_s = np.sum(wgt)

    # Calculate the shift and the total weight as a cumulative sum of the
    # weighted average of the two curves
    shift = np.sum((dt1[~mask] - dt2[~mask]) * wgt) / wgt_s
    comb_corr = np.sum((corr1[~mask] + corr2[~mask]) * wgt) / wgt_s

    # This saves from returning masked values
    try:
        shift = shift.filled(np.nan)
    except Exception:
        pass

    try:
        comb_corr = comb_corr.filled(np.nan)
    except Exception:
        pass

    return comb_corr, shift


def _create_G(n_ref: int) -> np.ndarray:
    """ Create the G matrix for the multi-trace alignment
    """

    G = None
    for jj in range(n_ref):
        line = list(range(n_ref))
        tline = [i for i in line if i != jj]
        tG = np.zeros((len(tline), len(tline)))
        if jj > 0:
            tG[:, jj - 1] = -1
        for ii in range(len(tline)):
            if tline[ii] > 0:
                tG[ii, tline[ii] - 1] = 1
        if G is None:
            G = tG
        else:
            G = np.vstack((G, tG))
    return G


def estimate_reftr_shifts_from_dt_corr(
        multi_ref_panel: dict, return_sim_mat: bool = False) -> dict:
    """ Combine velocity-change measurements of the same data performed with
    different references to a single curve.

    For a set of velocity-change measurements performed with different
    references this function estimates the relative offsets between all pairs
    of the measurements as a weighted average of their difference with the
    function :py:class:`~miic.core.stretch_mod.est_shift_from_dt_corr`.
    A least squares solution in computed that combines the pairwise
    differences to a consistent set of reference shifts. These shifts should
    be similar to the velocity variations measured between the reference
    traces. The consistent set of reference shifts is used to correct i.e.
    shift the similarity matricies to a common reference. Finally the
    corrected similarity matrices are averaged resulting in a single matrix
    that is interpreted as before. The position of the crest is the combined
    velocity change and the height of the crest is the correlation value.

    :type multi_ref_panel: dictionay
    :param multi_ref_panel: It is a dictionary with one (key,value) pair
        for each reference trace. Its structure is described
        in :func:`~seismic.monitor.stretch_mod.multi_ref_vchange`
    :type return_sim_mat: bool
    :param return_sim_mat: If `True` the returning dictionary contains also the
        similarity matrix `sim_mat`.
    :rtype: dict
    :return: **dv**: Dictionary with the following keys
        - *corr*: 2d ndarray containing the correlation value for the best
        match for each row of `mat` and for each time window.
        Its shape is: (len(tw),mat.shape[1])
        - *value*: 2d ndarray containing the stretch amount corresponding to
        the best match for each row of `mat` and for each time window.
        Stretch is a relative value corresponding to the negative relative
        velocity change -dv/v.
        Its shape is: (len(tw),mat.shape[1])
        - *sim_mat*: 3d array containing the similarity matricies that
        indicate the correlation coefficient with the reference for the
        different time windows, different times and different amount of
        stretching.
        Its shape is: (len(tw),mat.shape[1],len(strvec))
        - *second_axis*: It contains the stretch vector used for the velocity
        change estimate.
        - *value_type*: It is equal to stretch and specify the content of
        the returned value.
        - *method*: It is equal to single_ref and specify in which way the
        values have been obtained.
    """

    # Vector with the stretching amount
    stretch_vect = multi_ref_panel['reftr_0']['second_axis']
    delta = stretch_vect[1] - stretch_vect[0]

    n_ref = len(list(multi_ref_panel.keys()))

    corr = []
    shift = []

    if n_ref > 1:

        # Loop over reftr
        for reftr1 in np.sort(list(multi_ref_panel.keys())):
            ref_idx = [reftr for reftr in np.sort(list(multi_ref_panel.keys()))
                       if reftr != reftr1]
            for reftr2 in ref_idx:
                ccorr, sshift = est_shift_from_dt_corr(
                    np.squeeze(multi_ref_panel[reftr1]['value']),
                    np.squeeze(multi_ref_panel[reftr2]['value']),
                    np.squeeze(multi_ref_panel[reftr1]['corr']),
                    np.squeeze(multi_ref_panel[reftr2]['corr']))
                corr.append(ccorr)
                shift.append(sshift)

        G = _create_G(len(list(multi_ref_panel.keys())))
        W = np.diag(np.array(corr))
        D = np.array(shift)

        # This conversion is necessary to achive the same speed with diffrent
        # .dot implementation on different numpy versions.
        D = np.nan_to_num(D)
        W = np.nan_to_num(W)

        GTW = np.dot(G.T, W)

        # to_invert = np.dot(G.T, np.dot(W, G))
        to_invert = np.dot(GTW, G)

        # TODO: Get rid of the matrix inversion
        left_op = np.linalg.pinv(to_invert)

        # This conversion is necessary to achive the same speed with diffrent
        # .dot implementation on different numpy versions.
        left_op = np.nan_to_num(left_op)

        m = np.dot(left_op, np.dot(GTW, D))
        m = np.hstack((0, m))
        m = m - np.mean(m)

        # How many samples (int) each sim matrix must be rolled
        # PM October 2021 - require works the dtype casting in first
        # step caused errors, should be fine now?
        m = np.around(m / delta)  # , out=np.zeros_like(m, dtype='int32'))
        m = np.require(m, dtype='int32')

        row, col = np.squeeze(multi_ref_panel['reftr_0']['sim_mat']).shape

        stmp = np.zeros((row, col, n_ref))

        # Roll the sim_mat
        for (i, reftr) in enumerate(np.sort(list(multi_ref_panel.keys()))):

            stmp[:, :, i] = \
                np.roll(np.squeeze(multi_ref_panel[reftr]['sim_mat']),
                        m[i],
                        axis=1)

        # Create a msked array to evaluate the mean
        stmp_masked = np.ma.masked_array(
            stmp, mask=(np.isnan(stmp) | np.isinf(stmp)), fill_value=0)

        # Evaluate the sim_mat for the multi-ref approach as the mean
        # of the rolled sim_mat corresponfing to the individual reference
        # traces
        # When operating with masked arrays this operation creates a new
        # object so the default fill_values returns to be 1e20
        bsimmat = np.mean(stmp_masked, axis=2)

        corr = np.max(bsimmat, axis=1)
        dt = np.argmax(bsimmat, axis=1)

        dt = stretch_vect[dt]

        # Remove the mask
        bsimmat = bsimmat.filled(np.nan)

        try:
            corr = corr.filled(np.nan)
        except Exception:
            pass

        try:
            dt = dt.filled(np.nan).astype('int')
        except Exception:
            pass

        ret_dict = {'corr': corr,
                    'value': dt,
                    'second_axis': stretch_vect,
                    'value_type': 'stretch',
                    'method': 'multi_ref'}

        if return_sim_mat:
            ret_dict.update({'sim_mat': bsimmat})

        return ret_dict

    else:
        print("For a single reference trace use the appropriate function")
        return None


def multi_ref_vchange_and_align(
    corr_data: np.ndarray, ref_trs: np.ndarray, tw: np.ndarray = None,
    stretch_range: float = 0.1, stretch_steps: int = 100,
    sides: str = 'both', return_sim_mat: bool = False,
    remove_nans: bool = True,
    ref_tr_trim: Optional[Tuple[float, float]] = None,
        ref_tr_stats=None) -> dict:
    """ Multi-reference dv estimate and alignment

    :type corr_data: :class:`~numpy.ndarray`
    :param corr_data: 2d ndarray containing the correlation functions.
        One for each row.
    :type ref_trc: :class:`~numpy.ndarray`
    :param ref_trc: 1D array containing the reference trace to be shifted and
        compared to the individual traces in ``mat``
    :type tw: list of :class:`~numpy.ndarray` of int
    :param tw: list of 1D ndarrays holding the indices of sampels in the time
        windows to be use in the time shift estimate. The sampels are counted
        from the zero lag time with the index of the first sample being 0. If
        ``tw = None`` the full time range is used.
    :type stretch_range: scalar
    :param stretch_range: Maximum amount of relative stretching.
        Stretching and compression is tested from ``-stretch_range`` to
        ``stretch_range``.
    :type stretch_steps: scalar`
    :param stretch_steps: Number of shifted version to be tested. The
        increment will be ``(2 * stretch_range) / stretch_steps``
    :type single_sided: bool
    :param single_sided: if True zero lag time is on the first sample. If
        False the zero lag time is in the center of the traces.
    :type sides: str
    :param sides: Side of the reference matrix to be used for the stretching
        estimate ('both' | 'left' | 'right' | 'single') ``single`` is used for
        one-sided signals from active sources with zero lag time is on the
        first sample. Other options assume that the zero lag time is in the
        center of the traces.
    :type return_sim_mat: bool
    :param return_sim_mat: If `True` the returning dictionary contains also the
        similarity matrix `sim_mat`.
    :type remove_nans: bool
    :param remove_nans: If `True` applay :func:`~numpy.nan_to_num` to the
        given correlation matrix before any other operation.

    :rtype: Dictionary
    :return: **dv**: Dictionary with the following keys

        *corr*: 2d ndarray containing the correlation value for the best
            match for each row of ``mat`` and for each time window.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *value*: 2d ndarray containing the stretch amount corresponding to
            the best match for each row of ``mat`` and for each time window.
            Stretch is a relative value corresponding to the negative relative
            velocity change -dv/v.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *sim_mat*: 3d ndarray containing the similarity matricies that
            indicate the correlation coefficient with the reference for the
            different time windows, different times and different amount of
            stretching.
            Its dimension is: :py:func:`(len(tw),mat.shape[1],len(strvec))`
        *second_axis*: It contains the stretch vector used for the velocity
            change estimate.
        *vale_type*: It is equal to 'stretch' and specify the content of
            the returned 'value'.
        *method*: It is equal to 'single_ref' and specify in which "way" the
            values have been obtained.
    """

    # FIX: Trace back this problem to the original source and remove
    # this statement
    if remove_nans:
        corr_data = np.nan_to_num(corr_data)
        ref_trs = np.nan_to_num(ref_trs)

    if tw is not None and len(tw) > 1:
        print(" The multi-reference vchange evaluation doesn't handle multiple\
                time windows. Only the first time-window will be used")
        tw = tw[0]

    multi_ref_panel = multi_ref_vchange(
        corr_data, ref_trs, tw=tw, stretch_range=stretch_range,
        stretch_steps=stretch_steps, sides=sides, remove_nans=remove_nans,
        ref_tr_trim=ref_tr_trim, ref_tr_stats=ref_tr_stats)

    n_ref = len(list(multi_ref_panel.keys()))

    if n_ref > 1:
        dv = estimate_reftr_shifts_from_dt_corr(
            multi_ref_panel, return_sim_mat=return_sim_mat)
    else:
        # changed key here
        dv = multi_ref_panel['reftr_0']

    return dv


def time_shift_estimate(
    corr_data: np.ndarray, ref_trc: np.ndarray = None,
    tw: List[np.ndarray] = None, shift_range: float = 10,
    shift_steps: int = 100, single_sided: bool = False,
        return_sim_mat: bool = True, remove_nans: bool = True) -> dict:
    """ Time shift estimate through shifting and comparison.

    This function is intended to estimate shift of traces as they can occur
    in noise cross-correlation in case of drifting clocks.

    Time shifts are estimated comparing each correlation function stored
    in the ``corr_data`` matrix (one for each row) with ``shift_steps``
    shifted versions  of reference trace stored in ``ref_trc``.
    The maximum amount of shifting may be passed in ``shift_range``.
    The best match (shifting amount and corresponding correlation value) is
    calculated on different time windows. If ``tw = None`` the shifting is
    estimated on the whole trace.

    :type corr_data: :class:`~numpy.ndarray`
    :param corr_data: 2d ndarray containing the correlation functions.
        One for each row.
    :type ref_trc: :class:`~numpy.ndarray`
    :param ref_trc: 1D array containing the reference trace to be shifted and
        compared to the individual traces in ``mat``
    :type tw: list of :class:`~numpy.ndarray` of int
    :param tw: list of 1D ndarrays holding the indices of sampels in the time
        windows to be use in the time shift estimate. The sampels are counted
        from the zero lag time with the index of the first sample being 0. If
        ``tw = None`` the full time range is used.
    :type shift_range: scalar
    :param shift_range: Maximum amount of time shift in samples (in one
        direction).
        Shifting is tested in both directions from ``-shift_range`` to
        ``shift_range``
    :type shift_steps: scalar`
    :param shift_steps: Number of shifted version to be tested. The increment
            will be ``(2 * shift_range) / shift_steps``
    :type sinlge_sided: boolean
    :param single_sided: If ``True`` the zero lag time of the traces is in the
        first sample. If ``False`` zero lag is assumed to be in the center of
        the traces and the shifting is evaluated on the causal and acausal
        parts of the traces separately and averaged. This is done to avoid bias
        from velocity changes (stretching) in the case of strongly asymmetric
        traces.
    :type remove_nans: bool
    :param remove_nans: If `True` applay :func:`~numpy.nan_to_num` to the
        given correlation matrix before any other operation.

    :rtype: Dictionary
    :return: **dt**: Dictionary with the following keys

        *corr*: :class:`~numpy.ndarray` 2d ndarray containing the correlation
            value for the best match for each row of ``mat`` and for each
            time window.
            Its dimension is: :func:(len(tw),mat.shape[1])
        *value*: :class:`~numpy.ndarray` 2d ndarray containing the amount of
            shifting corresponding to the best match for each row of ``mat``
            and for each time window. Shift is measured in units of the
            sampling interval.
            Its dimension is: :py:func:`(len(tw),mat.shape[1])`
        *sim_mat*: 2d ndarray containing the similarity matrix that
            indicate the time shift respect to the reference for the
            selected time windows, for different times and different amount of
            shift.
            Its dimension is: :py:func:`(mat.shape[1],shift_steps)`
        *second_axis*: It contains the shift vector used for the velocity
            change estimate.
        *vale_type*: It is equal to 'shift' and specify the content of
            the returned 'value'.
        *method*: It is equal to 'time_shift' and specify in which "way" the
            values have been obtained.
    """

    # Mat must be a 2d vector in every case so
    mat = np.atleast_2d(corr_data)

    # FIX: Trace back this problem to the original source and remove
    # this statement
    if remove_nans:
        mat = np.nan_to_num(mat)

    # generate the reference trace if not given (use the whole time span)
    if ref_trc is None:
        ref_trc = np.nansum(mat, axis=0) / mat.shape[0]

    # generate time window if not given (use the full length of
    # the correlation trace)
    if tw is None:
        if single_sided:
            tw = time_windows_creation([0], [mat.shape[1]])
        # if center is in the middle use only half the length for the
        # time window
        else:
            tw = time_windows_creation(
                [0], [int(np.floor(mat.shape[1] / 2.))])

    # # taper and extend the reference trace to avoid interpolation
    # # artefacts at the ends of the trace
    # taper = cosine_taper(len(ref_trc), 0.05)
    # ref_trc *= taper

    # different values of shifting to be tested
    shifts = np.linspace(-shift_range, shift_range, shift_steps)

    # time axis
    if single_sided:
        time_idx = np.arange(len(ref_trc)) - (len(ref_trc) - 1.) / 2.
    else:
        time_idx = np.arange(len(ref_trc))

    # create the array to hold the shifted traces
    ref_shift = np.zeros((len(shifts), len(ref_trc)))

    # create a spline object for the reference trace
    ref_tr_spline = UnivariateSpline(time_idx, ref_trc, s=0, ext='const')

    # evaluate the spline object at different points and put in the prepared
    # array
    for (k, this_shift) in enumerate(shifts):
        ref_shift[k, :] = ref_tr_spline(time_idx - this_shift)

    # search best fit of the crosscorrs to one of the shifted ref_traces
    if single_sided:
        vdict = velocity_change_estimate(
            mat, tw, ref_shift, shifts, sides='single', return_sim_mat=True,
            remove_nans=remove_nans)
        corr = vdict['corr']
        shift = vdict['value']
        sim_mat = vdict['sim_mat']

    else:
        dtdict = velocity_change_estimate(
            mat, tw, ref_shift, shifts, sides='both', return_sim_mat=True,
            remove_nans=remove_nans)
        corr = dtdict['corr']
        shift = dtdict['value']
        sim_mat = dtdict['sim_mat']

    # create the result dictionary
    dt = {'corr': np.squeeze(corr),
          'value': np.squeeze(shift),
          'second_axis': shifts,
          'value_type': np.array(['shift']),
          'method': np.array(['time_shift'])}

    if return_sim_mat:
        dt.update({'sim_mat': np.squeeze(sim_mat)})

    return dt


def time_shift_apply(
    corr_data: np.ndarray, shift: np.ndarray,
        single_sided: bool = False) -> np.ndarray:
    """
    Correct for clock drifts that are given in shifts.

    :param corr_data: Matrix with one correlation trace per row
    :type corr_data: np.ndarray
    :param shift: 1 or 2D array holding shifts in number of samples compared to
        a reference that is assumed to be accurate. Hence, the input CorrStream
        will be shifted by -shift.
        If 1D: each correlation trace will be shifted by a scalar value.
        If 2D: Shape has to be equal to corr_data. The clock drift varies
        in the CorrTrace (i..e, shift is not constant anymore).
    :type shift: np.ndarray
    :param single_sided: corr_data contains only causal side, defaults to False
    :type single_sided: bool, optional
    :return: The shifted correlation matrix
    :rtype: np.ndarray
    """
    # Mat must be a 2d vector in every case so
    mat = np.atleast_2d(corr_data)
    # shift input
    # stretch is just a 1d array
    if len(shift.shape) == 1:
        t_shift = np.zeros([shift.shape[0], 1])
        t_shift[:, 0] = shift
        shift = t_shift
    # shift has the wrong length
    elif shift.shape[0] != mat.shape[0]:
        raise ValueError('shift.shape[0] must be equal corr_data.shape[0]')

    # if there is a significant clock drift during a correlation
    if shift.shape[1] > 1 and shift.shape[1] != mat.shape[1]:
        raise ValueError('shift.shape[1] must be equal corr_data.shape[1]')

    # time axis
    if single_sided:
        time_idx = np.arange(mat.shape[1])
    else:
        time_idx = np.arange(mat.shape[1]) - (mat.shape[1] - 1.) / 2.

    # allocate space for the result
    shifted_mat = np.zeros_like(mat)

    # stretch every line
    for ii, (ctr, delta) in enumerate(zip(mat, shift)):
        s = interp1d(
            time_idx, ctr, kind='linear', bounds_error=False, fill_value=0)
        shifted_mat[ii, :] = s(time_idx - delta)

    return shifted_mat


def time_stretch_apply(
    corr_data: np.ndarray, stretch: np.ndarray,
        single_sided: bool = False) -> np.ndarray:
    """ Apply time axis stretch to traces.

    Stretch the time axis of traces e.g. to compensate a velocity shift in the
    propagation medium.
    Such shifts can occur in corrlation traces in case of a drifting clock.
    This function applies the stretches. To correct for stretching
    estimated with :func:`~seismic.monitor.stretch_mod.time_stretch_estimate`
    you need to apply negative stretching.

    :type corr_data: :class:`~numpy.ndarray`
    :param corr_data: 2d ndarray containing the correlation functions that are
        to be shifted.
        One for each row.
    :type stretch: :class:`~numpy.ndarray`
    :param stretch: ndarray with stretch.shape[0] = corr_data.shape[0]
        containing the stretches relative units.

    :rtype: :class:`~numpy.ndarray`
    :return: **stretched_mat**: stretched version of the input matrix
    """

    # Mat must be a 2d vector in every case so
    mat = np.atleast_2d(corr_data)
    # check input
    # stretch is just a 1d array
    if len(stretch.shape) == 1:
        t_stretch = np.zeros([stretch.shape[0], 1])
        t_stretch[:, 0] = stretch
        stretch = t_stretch
    # stretch has the wrong length
    elif stretch.shape[0] != mat.shape[0]:
        raise ValueError('stretch.shape[0] must be equal corr_data.shape[0]')

    # shift has multiple columns (multiple measurements for the same time)
    if stretch.shape[1] > 1:
        stretch = np.delete(stretch, np.arange(1, stretch.shape[1]), axis=1)

    # taper and extend the reference trace to avoid interpolation
    # artefacts at the ends of the trace
    taper = cosine_taper(mat.shape[1], 0.05)
    mat *= np.tile(taper, [mat.shape[0], 1])

    # time axis
    if single_sided:
        time_idx = np.arange(mat.shape[1])
    else:
        time_idx = np.arange(mat.shape[1]) - (mat.shape[1] - 1.) / 2.

    # allocate space for the result
    stretched_mat = np.zeros_like(mat)

    # stretch every line
    for (ii, line) in enumerate(mat):
        s = UnivariateSpline(time_idx, line, s=0)
        stretched_mat[ii, :] = s(time_idx * (1 + stretch[ii]))
        stretched_mat[ii, :] = s(time_idx * np.exp(stretch[ii]))

    return stretched_mat


def create_shifted_ref_mat(
    ref_trc: np.ndarray, stats: CorrStats,
        shifts: np.ndarray) -> np.array:
    """
    Create a matrix of shifted versions of a reference trace.
    """
    assert len(ref_trc.shape) == 1 or ref_trc.shape[0] == 1, \
        f"'ref_trc' must be a 1-dimensional array,  has shape {ref_trc.shape}."
    # squeeze the trace to 1D array
    ref_trc = np.squeeze(ref_trc)
    # allocate space for the shifted traces
    ref_mat = np.zeros((len(shifts), len(ref_trc)))
    # create the time vector
    times = np.linspace(stats.start_lag, stats.end_lag, stats.npts)
    # create the spline
    s = UnivariateSpline(times, ref_trc, s=0)
    for ii, shift in enumerate(shifts):
        ref_mat[ii, :] = s(times + shift)

    return ref_mat


def compare_with_modified_reference(
    data: np.ndarray, ref_mat: np.ndarray,
        indices: np.ndarray) -> np.ndarray:
    """
    Compare a correlation matrix with a modified references.
    """
    # create mask for time window
    mat_mask = np.zeros_like(data)
    mat_mask[:, indices] = 1
    ref_mask = np.zeros_like(ref_mat)
    ref_mask[:, indices] = 1
    # mask array
    first = data[:, indices]
    second = ref_mat[:, indices]
    # correlation via dot product
    dprod = np.dot(first, second.T)
    # Normalization
    f_sq = np.sum(first ** 2, axis=1)
    s_sq = np.sum(second ** 2, axis=1)
    f_sq = f_sq.reshape(1, len(f_sq))
    s_sq = s_sq.reshape(1, len(s_sq))
    den = np.sqrt(np.dot(f_sq.T, s_sq))
    # apply normalization
    sim_mat = dprod / den

    return sim_mat



'''
Module that contains functions for preprocessing on obspy streams

:copyright:
    The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Tuesday, 20th July 2021 03:47:00 pm
Last Modified: Thursday, 29th June 2023 11:06:00 am
'''
from typing import List
from warnings import warn
from copy import deepcopy

import numpy as np
from obspy import Stream, Trace, UTCDateTime


def cos_taper_st(
    st: Stream, taper_len: float, lossless: bool,
        taper_at_masked: bool = False) -> Stream:
    """
    Applies a cosine taper to the input Stream.

    :param tr: Input Stream
    :type tr: :class:`~obspy.core.stream.Stream`
    :param taper_len: Length of the taper per side
    :type taper_len: float
    :param taper_at_masked: applies a split to each trace and merges again
        afterwards
    :type taper_at_masked: bool
    :param lossless: Lossless tapering pads the trace's ends with a copy of
        the trace's data before tapering. Note that you will want to trim
        the trace later to remove this artificial ends.
    type lossless: bool
    :return: Tapered Stream
    :rtype: :class:`~obspy.core.stream.Stream`

    .. note::
        This action is performed in place. If you want to keep the
        original data use :func:`~obspy.core.stream.Stream.copy`.
    """
    if isinstance(st, Trace):
        st = Stream([st])
    for ii, _ in enumerate(st):
        try:
            st[ii] = cos_taper(st[ii], taper_len, taper_at_masked, lossless)
        except ValueError as e:
            warn('%s, corresponding trace not tapered.' % e)
    return st


def cos_taper(
    tr: Trace, taper_len: float, taper_at_masked: bool,
        lossless: bool) -> Trace:
    """
    Applies a cosine taper to the input trace.

    :param tr: Input Trace
    :type tr: Trace
    :param taper_len: Length of the taper per side in seconds
    :type taper_len: float
    :param taper_at_masked: applies a split to each trace and merges again
        afterwards. Lossless tapering is not supporting if the trace contains
        masked values in the middle of the trace
    :type taper_at_masked: bool
    :param lossless: Lossless tapering pads the trace's ends with a copy of
        the trace's data before tapering. Note that you will want to trim
        the trace later to remove this artificial ends.
    type lossless: bool
    :return: Tapered Trace
    :rtype: Trace

    .. note::
        This action is performed in place. If you want to keep the
        original data use :func:`~obspy.core.trace.Trace.copy`.
    """
    if taper_len <= 0:
        raise ValueError('Taper length must be larger than 0 s')
    if taper_at_masked:
        if lossless:
            warn(
                'Tapering lossless at masked values is not supported')
        st = tr.split()
        st = cos_taper_st(st, taper_len, False, False)
        st = st.merge()
        if st.count():
            tr.data = st[0].data
            return tr
        else:
            raise ValueError('Taper length must be larger than 0 s')
    tl_n = round(taper_len*tr.stats.sampling_rate)
    if tl_n * 2 > tr.stats.npts:
        raise ValueError(
            'Taper Length * 2 has to be smaller or equal to trace\'s length.')
    if lossless:
        # pad trace
        tr.trim(
            starttime=tr.stats.starttime-taper_len,
            endtime=tr.stats.endtime+taper_len,
            pad=True, fill_value=0)
        # Change Fill value to data from the trace
        tr.data[:tl_n] = deepcopy(tr.data[tl_n:2*tl_n])
        tr.data[-tl_n:] = deepcopy(tr.data[-2*tl_n:-tl_n])
    taper = np.ones_like(tr.data)
    tap = np.sin(np.linspace(0, np.pi, tl_n*2))
    taper[:tl_n] = tap[:tl_n]
    taper[-tl_n:] = tap[-tl_n:]
    tr.data = np.multiply(tr.data, taper)
    return tr


def detrend_st(st: Stream, *args, **kwargs) -> Stream:
    """
    Detrends a stream while dealing with data gaps

    :param st: input Stream
    :type st: Stream
    :return: the obspy Stream detrended
    :rtype: Stream

    .. note::
        This action is performed in place. If you want to keep the
        original data use :func:`~obspy.core.stream.Stream.copy`.

    .. seealso:
        For accepted parameters consult the documentation of
        :func:`obspy.core.trace.Trace.detrend`
    """
    for tr in st:
        sst = tr.split()
        sst.detrend(*args, **kwargs)
        tr.data = sst.merge()[0].data
    return st


def stream_filter(st: Stream, ftype: str, filter_option: dict) -> Stream:
    """ Filter each trace of a Stream according to the given parameters

    This faction apply the specified filter function to all the traces in the
    present in the input :py:class:`~obspy.core.stream.Stream`.

    :type ftype: str
    :param ftype: String that specifies which filter is applied (e.g.
            ``"bandpass"``). See the `Supported Filter`_ section below for
            further details.
    :type filter_option: dict
    :param filter_option: Necessary arguments for the respective filter
        that will be passed on. (e.g. ``freqmin=1.0``, ``freqmax=20.0`` for
        ``"bandpass"``)
    :type parallel: bool (Default: True)
    :pram parallel: If the filtering will be run in parallel or not
    :type processes: int
    :pram processes: Number of processes to start (if None it will be equal
        to the number of cores available in the hosting machine)

    .. note::

        This operation is performed in place on the actual data arrays. The
        raw data is not accessible anymore afterwards. To keep your
        original data, use :func:`~miic.core.alpha_mod.stream_copy` to create
        a copy of your stream object.
        This function can also work in parallel an all or a specified number of
        cores available in the hosting machine.

    .. rubric:: _`Supported Filter`

    ``'bandpass'``
        Butterworth-Bandpass (uses :func:`obspy.signal.filter.bandpass`).

    ``'bandstop'``
        Butterworth-Bandstop (uses :func:`obspy.signal.filter.bandstop`).

    ``'lowpass'``
        Butterworth-Lowpass (uses :func:`obspy.signal.filter.lowpass`).

    ``'highpass'``
        Butterworth-Highpass (uses :func:`obspy.signal.filter.highpass`).

    ``'lowpassCheby2'``
        Cheby2-Lowpass (uses :func:`obspy.signal.filter.lowpassCheby2`).

    ``'lowpassFIR'`` (experimental)
        FIR-Lowpass (uses :func:`obspy.signal.filter.lowpassFIR`).

    ``'remezFIR'`` (experimental)
        Minimax optimal bandpass using Remez algorithm (uses
        :func:`obspy.signal.filter.remezFIR`).

    """
    if isinstance(st, Trace):
        st = Stream([st])
    if not isinstance(st, Stream):
        raise TypeError("'st' must be a 'obspy.core.stream.Stream' object")

    fparam = dict(
        [(kw_filed, filter_option[kw_filed]) for kw_filed in filter_option])

    # take care of masked traces and keep their order in the stream
    fst = Stream()
    for tr in st:
        sptr = tr.split()
        sptr.filter(ftype, **fparam)
        sptr.merge()
        fst += sptr
    st = fst

    # Change the name to help blockcanvas readability
    st_filtered = st
    return st_filtered


def stream_mask_at_utc(
    st: Stream, starts: List[UTCDateTime], ends: List[UTCDateTime] = None,
        masklen: float = None, reverse: bool = False) -> Stream:
    """
    Mask the Data in the Stream between the times given by ``starts`` and
    ``ends`` or between ``starts`` and ``starts``+``masklen``.

    :param st: Input Strem to be tapered
    :type st: Stream
    :param starts: Start-time (in UTC) that the masked values should start from
    :type starts: List[UTCDateTime]
    :param ends: End times (in UTC) of the masked values. Has to have the same
        length as starts. If None, `masklen` has to be defined,
        defaults to None.
    :type ends: List[UTCDateTime], optional
    :param masklen: Alternatively to providing ends, one can provide a constant
        length (in s) per mask, defaults to None.
    :type masklen: float, optional
    :param reverse: Only keep the data in the mask unmasked and mask everything
        else. Defaults to False.
    :type reverse: bool, optional
    :raises ValueError: If `ends`, `starts`, and `masklen` are incompatible
        with each other.

    .. warning:: This function will not taper before and after the mask.
        If you should desire to do so use.
        :func:`~seismic.correlate.preprocessing_stream.cos_taper_st` and set
        ``taper_at_mask`` to ``True``.
    """
    msg = 'Provide either the length of the mask or a list of ends with '\
        + 'identical length as the list of starts.'
    if masklen is None and ends is None:
        raise ValueError(msg)
    elif masklen is not None and ends is not None:
        raise ValueError(msg)
    elif ends is not None and len(ends) != len(starts):
        raise ValueError('Ends must have the same length as starts.')
    starts = np.array(starts)
    if ends is None:
        ends = starts + masklen
    else:
        ends = np.array(ends)
    for tr in st:
        trace_mask_at_utc(tr, starts, ends, reverse)
    return st


def trace_mask_at_utc(
    tr: Trace, starts: np.ndarray,
        ends: np.ndarray, reverse: bool):
    """
    .. seealso::
        :func:`~seismic.correlate.preprocessing_stream.stream_mask_at_utc`
    """
    start = tr.stats.starttime
    end = tr.stats.endtime
    # starts in trace
    ii = (starts < end) * (starts > start)
    # ends in trace
    jj = (ends < end) * (ends > start)

    mask = np.zeros(tr.data.shape, dtype=bool)

    # masks that are completlely in trace
    kk = jj*ii
    for s, e in zip(starts[kk], ends[kk]):
        # Find start-index in trace
        t = s-start
        ns = int(np.floor(tr.stats.sampling_rate*t))
        # find stop index
        t = e-start
        ne = int(np.ceil(tr.stats.sampling_rate*t))+1
        mask[ns:ne] = True

    # only start of mask in trace
    ll = ii * ~jj
    for s in starts[ll]:
        # Find start-index in trace
        t = s-start
        ns = int(np.floor(tr.stats.sampling_rate*t))
        mask[ns:] = True

    # only end of mask in trace
    ll = ~ii * jj
    for e in ends[ll]:
        # Find start-index in trace
        t = e-start
        ne = int(np.ceil(tr.stats.sampling_rate*t))+1
        mask[:ne] = True

    if reverse:
        mask = ~mask
    # Mask the array
    tr.data = np.ma.array(tr.data, mask=mask, hard_mask=True, fill_value=0)
    # Hard mask saves RAM as the data will essentially be discarded



'''
:copyright:
   The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Tuesday, 15th June 2021 04:12:18 pm
Last Modified: Friday, 16th June 2023 03:54:21 pm
'''

from datetime import datetime
from glob import glob
from typing import List, Tuple, Optional
import warnings
from zipfile import BadZipFile

import matplotlib.pyplot as plt
import numpy as np

from seismic.plot.plot_dv import plot_dv
from seismic.utils import miic_utils as mu
from seismic.correlate.stats import CorrStats


class DV(object):
    """
    A simple object to save velocity changes.
    """
    def __init__(
        self, corr: np.ndarray, value: np.ndarray, value_type: str,
        sim_mat: np.ndarray, second_axis: np.ndarray, method: str,
        stats: CorrStats, stretches: np.ndarray = None,
        corrs: np.ndarray = None, n_stat: np.ndarray = None,
            dv_processing: dict = None):
        """
        Creates an object designed to hold and process velocity changes.

        :param corr: Array with maximum Correlation coefficients.
        :type corr: np.ndarray
        :param value: Array with decimal velocity changes / stretches.
        :type value: np.ndarray
        :param value_type: Type of value (e.g., stretch)
        :type value_type: str
        :param sim_mat: Similarity Matrix with the dimensions time and stretch.
        :type sim_mat: np.ndarray
        :param second_axis: second axis (e.g., the stretch axis)
        :type second_axis: np.ndarray
        :param method: Stretching method that was used
        :type method: str
        :param stats: Stats of the correlation object that was used
        :type stats: CorrStats
        :param stretches: Scatter of the value (e.g., stretch).
            In case this dv object holds an average
            of several dvs, the user can choose to keep all dvs' stretches.
            defaults to None
        :type stretches: np.ndarray, optional
        :param corrs: Scatter of the correlation coefficient (e.g., stretch).
            In case this dv object holds an average
            of several dvs, the user can choose to keep all dvs' corrs.
            defaults to None
        :type corrs: np.ndarray, optional
        :param n_stat: Number of stations used for the stack at corr_start t.
            Has the same shape as `value` and `corr`. Defaults to None
        :type n_stat: np.ndarray, optional
        """
        # Allocate attributes
        self.value_type = value_type
        self.corr = corr
        self.value = value
        self.sim_mat = sim_mat
        self.corrs = corrs
        self.stretches = stretches
        self.n_stat = n_stat
        self.second_axis = second_axis
        self.method = method
        self.stats = stats
        self.dv_processing = dv_processing
        # Available indices
        self.avail = ~np.isnan(self.corr)

    def __str__(self):
        """
        Print a prettier string.
        """
        code = f'{self.stats.network}.{self.stats.station}.'\
            + self.stats.channel
        out = f'{self.method} {self.value_type} velocity change estimate of '\
            + f'{code}.\nstarttdate: {min(self.stats.corr_start).ctime()}\n'\
            + f'enddate: {max(self.stats.corr_end).ctime()}\n\n'\
            + f'processed with the following parameters: {self.dv_processing}'
        if self.method == np.array(['time_shift']):
            out = f'Time shift estimate of {code}.\nstarttdate: '\
                + f'{min(self.stats.corr_start).ctime()}\nenddate: '\
                + f'{max(self.stats.corr_end).ctime()}'
        return out

    def save(self, path: str):
        """
        Saves the dv object to a compressed numpy binary format. The DV can
        later be read using :func:`~seismic.monitor.dv.read_dv`.

        :param path: output path
        :type path: str
        """
        kwargs = mu.save_header_to_np_array(self.stats)
        if self.dv_processing is not None:
            try:
                sides = self.dv_processing['sides']
            except KeyError:
                sides = 'unknown'
            kwargs.update({
                'freq_min': self.dv_processing['freq_min'],
                'freq_max': self.dv_processing['freq_max'],
                'tw_len': self.dv_processing['tw_len'],
                'tw_start': self.dv_processing['tw_start'],
                'sides': sides
            })
        kwargs.update({
            'method_array': np.array([self.method]),
            'vt_array': np.array([self.value_type]),
            'corr': self.corr,
            'value': self.value,
            'sim_mat': self.sim_mat,
            'second_axis': self.second_axis,
            'stretches': self.stretches,
            'corrs': self.corrs,
            'n_stat': self.n_stat
        })
        # Np load will otherwise be annoying
        to_save = {k: v for k, v in kwargs.items() if v is not None}
        np.savez_compressed(
            path, **to_save)

    def plot(
        self, save_dir: str = '.', figure_file_name: str = None,
        mark_time: datetime = None, normalize_simmat: bool = False,
        sim_mat_Clim: List[float] = [], xlim: Tuple[datetime, datetime] = None,
        ylim: Tuple[int, int] = None, plot_scatter: bool = False,
        figsize: Tuple[float, float] = (9, 11), dpi: int = 144,
        title: str = None, return_ax=False, style: str = 'technical',
        dateformat: str = '%d %b %y') -> Tuple[
            plt.figure, List[plt.axis]]:
        r"""
        Plots the dv object into a *multi-panel-view* of `similarity matrix`
        `dv-value`, and `correlation coefficient`.

        :param save_dir: Directory to save the figure to, defaults to '.'
        :type save_dir: str, optional
        :param figure_file_name: Filename of the figure. If None, the figure
            will not be save, defaults to None.
        :type figure_file_name: str, optional
        :param mark_time: times to put ticks to. If not set, they will be
            determined automatically, defaults to None.
        :type mark_time: datetime, optional
        :param normalize_simmat: Normalise the similarity matrix?
            Defaults to False.
        :type normalize_simmat: bool, optional
        :param sim_mat_Clim: Color limits for the similarity matrix
            , defaults to []. Format is [low_limit, upper_limit]
        :type sim_mat_Clim: List[float], optional
        :param xlim: Time Limits to plot, defaults to None.
        :type xlim: Tuple[datetime, datetime], optional
        :param ylim: Y-limits (e.g., stretch) of the plot, defaults to None
        :type ylim: Tuple[int, int], optional
        :param plot_scatter: If set to True, the scatter of all used
            combinations is plotted in form of a histogram.
            Defaults to False.
        :type plot_scatter: bool, optional
        :param figsize: Size of the figure/canvas, defaults to (9, 11)
        :type figsize: Tuple[float, float], optional
        :param dpi: Pixels per inch, defaults to 144
        :type dpi: int, optional
        :param title: Define a custom title. Otherwise, an automatic title
            will be created, defaults to None
        :type title: str, optional
        :param return_ax: Return plt.figure and list of axes.
            Defaults to False.
            This overwrites any choice to save the figure.
        :type return_ax: bool, optional
        :param style: Style of the plot. Defaults to 'technical'. The other
            option would be 'publication' (looks better but is less
            informative).
        :type style: str, optional
        :param dateformat: Format of the date on the x-axis. Defaults to
            '%d %b %y'.
        :type dateformat: str, optional
        :returns: If `return_ax` is set to True it returns fig and axes.
        """
        return plot_dv(
            style, self, save_dir=save_dir,
            figure_file_name=figure_file_name, mark_time=mark_time,
            normalize_simmat=normalize_simmat, sim_mat_Clim=sim_mat_Clim,
            figsize=figsize, dpi=dpi, xlim=xlim, ylim=ylim,
            title=title, plot_scatter=plot_scatter, return_ax=return_ax,
            dateformat=dateformat)

    def smooth_sim_mat(
            self, win_len: int, exclude_corr_below: Optional[float] = None):
        """
        Smoothes the similarity matrix along the correlation time axis.

        :param win_len: Length of the window in number of samples.
        :type win_len: int
        :param exclude_corr_below: Exclude data points with a correlation
            coefficient below the chosen threshold. Defaults to None (i.e.,
            include all)
        :type exclude_corr_below: float

        :note::

            This action is perfomed in-place and irreversible as the original
            data are not accesible any more.
        """
        if exclude_corr_below is not None:
            self.sim_mat[self.sim_mat < exclude_corr_below] = np.nan
        self.sim_mat = mu.nan_moving_av(self.sim_mat, int(win_len/2), axis=0)

        # Compute the dependencies again
        self.corr = np.nanmax(self.sim_mat, axis=1)
        self.value = self.second_axis[
            np.argmax(np.nan_to_num(self.sim_mat), axis=1)]
        return self


def read_dv(path: str) -> DV:
    """
    Reads a saved velocity change object from an **.npz** file.

    :param path: Path to file, wildcards allowed
    :type path: str
    :return: the corresponding and converted DV object
    :rtype: DV, List[DV] if path contains wildcards
    """
    if '*' in path or '?' in path:
        dvl = []
        for p in glob(path):
            try:
                dvl.append(read_dv(p))
            except BadZipFile:
                warnings.warn(f'File {p} corrupt, skipping..')
        if not len(dvl):
            raise FileNotFoundError(
                f'No files that adhere the pattern {path} found.')
        return dvl
    loaded = np.load(path)
    stats = CorrStats(mu.load_header_from_np_array(loaded))
    # to check that compatibility works
    vt = loaded['vt_array']
    while not isinstance(vt, str):
        vt = vt[0]
    method = loaded['method_array']
    while not isinstance(method, str):
        method = method[0]
    try:
        stretches = loaded['stretches']
        corrs = loaded['corrs']
    except KeyError:
        stretches = corrs = None
    try:
        n_stat = loaded['n_stat']
    except KeyError:
        n_stat = None
    try:
        dv_processing = dict(
            freq_min=float(loaded['freq_min']),
            freq_max=float(loaded['freq_max']),
            tw_start=float(loaded['tw_start']),
            tw_len=float(loaded['tw_len']))
    except KeyError:
        dv_processing = None
    return DV(
        loaded['corr'], loaded['value'], vt, loaded['sim_mat'],
        loaded['second_axis'], method, stats=stats, stretches=stretches,
        corrs=corrs, n_stat=n_stat, dv_processing=dv_processing)



'''
Module that contains functions for preprocessing in the time domain

:copyright:
    The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Tuesday, 20th July 2021 03:24:01 pm
Last Modified: Thursday, 29th June 2023 11:15:56 am
'''
from copy import deepcopy

import numpy as np
from scipy.fftpack import next_fast_len
from scipy import signal
from scipy.signal import detrend as sp_detrend
import obspy.signal as osignal

from seismic.utils.fetch_func_from_str import func_from_str


def clip(A: np.ndarray, args: dict, params: dict) -> np.ndarray:
    """
    Clip time series data at a multiple of the standard deviation

    Set amplitudes exeeding a certain threshold to this threshold.
    The threshold for clipping is estimated as the standard deviation of each
    trace times a factor specified in `args`.

    :Note: Traces should be demeaned before clipping.

    :type A: numpy.ndarray
    :param A: time series data with time oriented along the first \\
        dimension (columns)
    :type args: dictionary
    :param args: the only keyword allowed is `std_factor` describing the \\
        scaling of the standard deviation for the clipping threshold
    :type params: dictionary
    :param params: not used here

    :rtype: numpy.ndarray
    :return: clipped time series data
    """
    stds = np.nanstd(A, axis=1)
    for ind in range(A.shape[0]):
        ts = args['std_factor']*stds[ind]
        A[ind, A[ind, :] > ts] = ts
        A[ind, A[ind, :] < -ts] = -ts
    return A


def detrend(A: np.ndarray, args: dict, params: dict) -> np.ndarray:
    """
    Detrend wrapper. Can call scipy.signal.detrend or a faster custom function.

    :param A: 1 or 2D array
    :type A: np.ndarray
    :param args: arguments for detrend function
    :type args: dict
    :param params: params file
    :type params: dict
    :raises ValueError: For unknown detrend method
    :raises ValueError: For unknown detrend type
    :return: detrended data
    :rtype: np.ndarray
    """
    method = args.get('method', 'qr')
    if method == 'scipy':
        return detrend_scipy(A, args, params)
    elif method == 'qr':
        detrend_type = args.get('type', 'linear')
        if detrend_type == 'linear':
            return detrendqr(A)
        elif detrend_type == 'constant':
            return demean(A)
        else:
            raise ValueError('Unknown detrend type')
    else:
        raise ValueError('Unknown detrend method')


def detrend_scipy(A: np.ndarray, args: dict, params: dict) -> np.ndarray:
    """
    Remove trend from data

    .. seealso:: :func:`scipy.signal.detrend` for input arguments
    """
    if np.all(np.isnan(A)):
        return A
    for ii in range(A.shape[0]):
        try:
            A[ii, ~np.isnan(A[ii])] = sp_detrend(
                A[ii, ~np.isnan(A[ii])], axis=-1, overwrite_data=True,
                **args)
        except ZeroDivisionError:
            # When the line is nothing but nans
            pass
    return A


def detrendqr(data: np.ndarray) -> np.ndarray:
    """
    Remove trend from data using QR decomposition. Faster than
    scipy.signal.detrend. Shamelessly adapted from
    NoisePy

    .. seealso:: Adapted from `NoisePy <https://doi.org/10.1785/0220190364>`_

    :param data: 1 or 2D array
    :type data: np.ndarray
    :raises ValueError: For data with more than 2 dimensions
    :return: detrended data
    :rtype: np.ndarray
    """
    npts = data.shape[-1]
    X = np.ones((npts, 2))
    X[:, 0] = np.arange(0, npts, dtype=np.float32) / npts
    Q, R = np.linalg.qr(X)
    rq = np.dot(np.linalg.inv(R), Q.transpose())
    if data.ndim == 1:
        coeff = np.dot(rq, data)
        data = data - np.dot(X, coeff)
    elif data.ndim == 2:
        for ii in range(data.shape[0]):
            coeff = np.dot(rq, data[ii])
            data[ii] = data[ii] - np.dot(X, coeff)
    else:
        raise ValueError('data must be 1D or 2D')
    return data


def demean(data: np.ndarray) -> np.ndarray:
    """
    Demean data

    :param data: 1 or 2D array
    :type data: np.ndarray
    :return: demeaned data
    :rtype: np.ndarray
    """
    return data - np.nanmean(data, axis=-1, keepdims=True)


def mute(A: np.ndarray, args: dict, params: dict) -> np.ndarray:
    """
    Mute parts of data that exceed a threshold

    To completely surpress the effect of data with high amplitudes e.g. after
    aftershocks these parts of the data are muted (set to zero). The respective
    parts of the signal are identified as those where the envelope in a given
    frequency exceeds a threshold given directly as absolute numer or as a
    multiple of the data's standard deviation. A taper of length `taper_len` is
    applied to smooth the edges of muted segments. Setting `extend_gaps` to
    True will ensure that the taper is applied outside the segments and data
    inside these segments will all zero. Edges of the data will be tapered too
    in this case.



    :type A: numpy.ndarray
    :param A: time series data with time oriented along the first dimension
        (columns)
    :type args: dictionary
    :param args: the following keywords are allowed:

        * `filter`:
            (dictionary) description of filter to be applied before
            calculation of the signal envelope. If not given the envelope is
            calculated from raw data. The value of the keyword filter is the
            same as the `args` for the function `TDfilter`.
        * `threshold`:
            (float) absolute amplitude of threshold for muting
        * `std_factor`:
            (float) alternativly to an absolute number the threhold
            can be estimated as a multiple of the standard deviation if the
            scaling is given in as value of the keyword `std_factor`. If
            neither `threshold` nor `std_factor` are given `std_factor=1` is
            assumed.
        * `extend_gaps`:
            (boolean) if True date above the threshold is
            guaranteed to be muted, otherwise tapering will leak into these
            parts. This step involves an additional convolution.
        * `taper_len`:
            (float) length of taper for muted segments in seconds

    :type params: dictionary
    :param params: filled automatically by `pxcorr`

    :rtype: numpy.ndarray
    :return: clipped time series data

    :Example:
        ``args={'filter':{'type':'bandpass', 'freqmin':1., 'freqmax':6.},
        'taper_len':1., 'threshold':1000, 'std_factor':1, 'extend_gaps':True}``
    """

    if args['taper_len'] == 0:
        raise ValueError('Taper Length cannot be zero.')

    # return zeros if length of traces is shorter than taper
    ntap = int(args['taper_len']*params['sampling_rate'])
    if A.shape[1] <= ntap:
        return np.zeros_like(A)

    # filter if asked to
    if 'filter' in list(args.keys()):
        C = TDfilter(A, args['filter'], params)
    else:
        C = A

    # calculate envelope
    D = np.abs(C)

    # calculate threshold
    if 'threshold' in list(args.keys()):
        thres = np.zeros(A.shape[0]) + args['threshold']
    elif 'std_factor' in list(args.keys()):
        thres = np.std(C, axis=1) * args['std_factor']
    else:
        thres = np.std(C, axis=1)

    # calculate mask
    mask = np.ones_like(D)
    mask[D > np.tile(thres, (D.shape[1], 1)).T] = 0
    # extend the muted segments to make sure the whole segment is zero after
    if args['extend_gaps']:
        tap = np.ones(ntap)/ntap
        for ind in range(A.shape[0]):
            mask[ind, :] = np.convolve(mask[ind, :], tap, mode='same')
        nmask = np.ones_like(D)
        nmask[mask < 1.] = 0
    else:
        nmask = mask

    # apply taper
    tap = 2. - (np.cos(np.arange(ntap, dtype=float)/ntap*2.*np.pi) + 1.)
    tap /= ntap
    for ind in range(A.shape[0]):
        nmask[ind, :] = np.convolve(nmask[ind, :], tap, mode='same')

    # mute data with tapered mask
    A *= nmask
    return A


def normalizeStandardDeviation(
        A: np.ndarray, args: dict, params: dict) -> np.ndarray:
    """
    Divide the time series by their standard deviation

    Divide the amplitudes of each trace by its standard deviation.

    :type A: numpy.ndarray
    :param A: time series data with time oriented along the first \\
        dimension (columns)
    :type args: dictionary
    :param args: not used here
    :type params: dictionary
    :param params: not used here

    :rtype: numpy.ndarray
    :return: normalized time series data
    """
    std = np.std(A, axis=-1)
    # avoid creating nans or Zerodivisionerror
    std[np.where(std == 0)] = 1
    A = (A.T/std).T
    return A


def signBitNormalization(
        A: np.ndarray, args: dict, params: dict) -> np.ndarray:
    """
    One bit normalization of time series data

    Return the sign of the samples (-1, 0, 1).

    :type A: numpy.ndarray
    :param A: time series data with time oriented along the first \\
        dimension (columns)
    :type args: dictionary
    :param args: not used here
    :type params: dictionary
    :param params: not used here

    :rtype: numpy.ndarray
    :return: 1-bit normalized time series data
    """
    return np.sign(A)


def taper(A: np.ndarray, args: dict, params: dict) -> np.ndarray:
    """
    Taper to the time series data

    Apply a simple taper to the time series data.

    `args` has the following structure:

        args = {'type':`type of taper`,taper_args}``

        `type` may be `cosine_taper` with the corresponding taper_args `p` the
        percentage of the traces to taper. Possibilities of `type` are \\
        given by `obspy.signal`.

        :Example:
            ``args = {'type':'cosine_taper','p':0.1}``

    :type A: numpy.ndarray
    :param A: time series data with time oriented along the first \\
        dimension (columns)
    :type args: dictionary
    :param args: arguments dictionary as described above
    :type params: dictionary
    :param params: not used here

    :rtype: numpy.ndarray
    :return: tapered time series data
    """
    if args['type'] == 'cosine_taper':
        func = osignal.invsim.cosine_taper
    else:
        func = getattr(signal, args['type'])
    args = deepcopy(args)
    args.pop('type')
    tap = func(A.shape[1], **args)
    A *= np.tile(np.atleast_2d(tap), (A.shape[0], 1))
    return A


def TDnormalization(A: np.ndarray, args: dict, params: dict) -> np.ndarray:
    """
    Amplitude dependent time domain normalization

    Calculate the envelope of the filtered trace, smooth it in a window of
    length `windowLength` and normalize the waveform by this trace. The two
    used keywords in `args` are `filter and `windowLength` that describe the
    filter and the length of the envelope smoothing window, respectively.

    `args` has the following structure:

        args = {'windowLength':`length of the envelope smoothing window in \\
        [s]`,'filter':{'type':`filterType`, fargs}}``

        `type` may be `bandpass` with the corresponding fargs `freqmin` and \\
        `freqmax` or `highpass`/`lowpass` with the `fargs` `freqmin`/`freqmax`

        :Example:
            ``args = {'windowLength':5,'filter':{'type':'bandpass',
            'freqmin':0.5, 'freqmax':2.}}``

    :type A: numpy.ndarray
    :param A: time series data with time oriented along the first \\
        dimension (columns)
    :type args: dictionary
    :param args: arguments dictionary as described above
    :type params: dictionary
    :param params: not used here

    :rtype: numpy.ndarray
    :return: normalized time series data
    """
    if args['windowLength'] <= 0:
        raise ValueError('Window Length has to be greater than 0.')
    # filter if args['filter']
    if args['filter']:
        B = TDfilter(A, args['filter'], params)
        # func = getattr(osignal, args['filter']['type'])
        # fargs = deepcopy(args['filter'])
        # fargs.pop('type')
        # B = func(A.T, df=params['sampling_rate'], **fargs).T
    else:
        B = deepcopy(A)
    # simple calculation of envelope
    B = B**2
    # smoothing of envelope in both directions to avoid a shift
    window = (
        np.ones(int(np.ceil(args['windowLength'] * params['sampling_rate'])))
        / np.ceil(args['windowLength']*params['sampling_rate']))
    if len(B.shape) == 1:
        B = np.convolve(B, window, mode='same')
        B = np.convolve(B[::-1], window, mode='same')[::-1]
        # damping factor
        B += np.max(B)*1e-6
    else:
        for ind in range(B.shape[0]):
            B[ind, :] = np.convolve(B[ind], window, mode='same')
            B[ind, :] = np.convolve(B[ind, ::-1], window, mode='same')[::-1]
            # damping factor
            B[ind, :] += np.max(B[ind, :])*1e-6
    # normalization
    A /= np.sqrt(B)
    return A


def TDfilter(A: np.ndarray, args: dict, params: dict) -> np.ndarray:
    """
    Filter time series data

    Filter in time domain. Types of filters are defined by `obspy.signal`.

    `args` has the following structure:

        args = {'type':`filterType`, fargs}

        `type` may be `bandpass` with the corresponding fargs `freqmin` and
        `freqmax` or `highpass`/`lowpass` with the `fargs` `freqmin`/`freqmax`

        :Example:
            ``args = {'type':'bandpass','freqmin':0.5,'freqmax':2.}``

    :type A: numpy.ndarray
    :param A: time series data with time oriented along the first \\
        dimension (columns)
    :type args: dictionary
    :param args: arguments dictionary as described above
    :type params: dictionary
    :param params: not used here

    :rtype: numpy.ndarray
    :return: filtered time series data
    """
    func = func_from_str('obspy.signal.filter.%s' % args['type'])
    args = deepcopy(args)
    args.pop('type')
    A = func(A, df=params['sampling_rate'], **args)
    return A


def zeroPadding(A: np.ndarray, args: dict, params: dict, axis=1) -> np.ndarray:
    """
    Append zeros to the traces

    Pad traces with zeros to increase the speed of the Fourier transforms and
    to avoid wrap around effects. Three possibilities for the length of the
    padding can be set in ``args['type']``:

    - ``nextFastLen``:
        Traces are padded to a length that is the next fast
        fft length
    - ``avoidWrapAround``:
        depending on length of the trace that is to be used
        the padded part is just long enough to avoid wrap around
    - ``avoidWrapFastLen``:
        Use the next fast length that avoids wrap around

    :Example: ``args = {'type':'avoidWrapPowerTwo'}``

    :type A: numpy.ndarray
    :param A: time series data with time oriented along the first \\
        dimension (columns)
    :type args: dictionary
    :param args: arguments dictionary as described above
    :type params: dictionary
    :param params: not used here
    :param axis: axis to pad on
    :type axis: tuple, optional

    :rtype: numpy.ndarray
    :return: zero padded time series data
    """
    if A.ndim > 2 or axis > 1:
        raise NotImplementedError('Only two-dimensional arrays are supported.')
    if A.ndim == 1 and len(A) == 0:
        raise ValueError('Input Array is empty')
    npts = A.shape[axis]
    if A.ndim == 2:
        ntrc = A.shape[axis-1]
    elif A.ndim == 1:
        ntrc = 1
    if args['type'] == 'nextFastLen':
        N = next_fast_len(npts)
    elif args['type'] == 'avoidWrapAround':
        N = npts + params['sampling_rate'] * params['lengthToSave']
    elif args['type'] == 'avoidWrapFastLen':
        N = next_fast_len(int(
            npts + params['sampling_rate'] * params['lengthToSave']))
    else:
        raise ValueError("type '%s' of zero padding not implemented" %
                         args['type'])

    if axis == 0:
        A = np.concatenate(
            (A, np.zeros((N-npts, ntrc), dtype=np.float32)), axis=axis)
    else:
        A = np.concatenate(
            (A, np.zeros((ntrc, N-npts), dtype=np.float32)), axis=axis)
    return A
